# -*- coding: utf-8 -*-
"""
Created on Wed Mar  9 17:06:44 2016

@author: Tobias Jachowski
"""
import collections
import cloudpickle
import inspect
import persistent
from ipywidgets import widgets
from IPython.display import display  # Used to display widgets in the notebook


class unboundfunction(persistent.Persistent):
    """
    Class to hold references to functions and still be able to pickle them.
    To reference the function you want to bind:
    function_reference = boundmethod(function)
    """
    def __init__(self, ft, **kwargs):
        self.ft = ft

    def __getstate__(self):
        return cloudpickle.dumps(self.ft)

    def __setstate__(self, cloudepickled_ft):
        self.ft = cloudpickle.loads(cloudepickled_ft)

    def __call__(self, *a, **kw):
        return self.ft(*a, **kw)


class boundmethod(persistent.Persistent):
    """
    Class to hold references to methods and still be able to pickle them.
    To reference the method you want to bind:
    method_reference = boundmethod(method)
    """
    def __init__(self, mt, **kwargs):
        self.mt = mt

    def __getstate__(self):
        return self.mt.__self__, self.mt.__func__.__name__

    def __setstate__(self, tuple_self_function_name):
        (s, fn) = tuple_self_function_name
        self.mt = getattr(s, fn)

    def __call__(self, *a, **kw):
        return self.mt(*a, **kw)


class Attributes(persistent.Persistent):
    """
    Store a dictionary in ZODB and use keys of the dict as attributes of an
    instance of Attributes.
    """
    def __init__(self):
        self._store = {}

    def values(self):
        return self._store.values()

    def __len__(self):
        return self._store.__len__()

    def __iter__(self):
        return self._store.__iter__()

    def __contains__(self, key):
        return key in self._store

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __delitem__(self, key):
        del self._store[key]

    def __getattr__(self, name):
        """
        Allow keys of status to be used as attributes
        """
        if name in self._store:
            return self._store[name]
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        """
        A __setattr__() hook will also override the Persistent __setattr__()
        hook.  User code must treat it much like __getattribute__().  The
        user-defined code must call _p_setattr() first to handle special
        attributes; _p_setattr() takes the attribute name and value.  If it
        returns True, Persistent handled the attribute. If not, the user code
        can run. If the user code modifies the object’s state, it must assigned
        to _p_changed
        """
        if not self._p_setattr(name, value):
            if '_store' in self.__dict__:
                self._store[name] = value
                self._p_changed = True  # inform ZODB of change
            else:
                super().__setattr__(name, value)


class FakeWidget(object):
    """
    FakeWidget is used for environments, where the graphical backend does not
    support instantiation of ipywidgets.widgets (self._create_widget()).

    Parameters
    ----------
    value : int, float, bool or None
      The value of the FakeWidget. If the value is set to None, a FakeWidget
      whose value is set to None and whose value cannot be changed anymore is
      created, to represent a "FakeButton" widget.
    description : str
      This value should describe what the value is supposed to represent.
    """
    def __init__(self, value, description, **kwargs):
        self._value = value
        self.description = description

    def __repr__(self):
        value = self.value
        if value is None:
            value = 'Button'
        return ''.join((self.description, ': %s' % value))

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        # prevent changing the value from/to None
        if self._value is not None and value is not None:
            self._value = value

# make relation of value type and widget class
WIDGET_CLASS = {
    None: widgets.Button,
    bool: widgets.Checkbox,
    float: widgets.FloatText,
    int: widgets.FloatText,
    list: widgets.SelectMultiple,
    tuple: widgets.SelectMultiple
}


class InteractiveAttributes(persistent.Persistent):
    """
    Stores widgets and implements methods to add widgets, change values of
    widgets, access values of widgets, and link widgets to callback functions
    called upon value change.
    """

    def __init__(self, **kwargs):
        # Widgets which graphically represent values and thereby add
        # interactive functionality to change and inspect values
        self._widgets = {}
        # self._widgets = AttributeDict()

        # List of callback functions [function1, function2, ...], which are
        # called whenever a value of a widget is changed
        self._callback_functions = {}

        # Controls whether callback functions upon change of widget value are
        # called or not.
        # Set to False, when you don't want the callback functions called.
        self._widget_callbacks_active = True

    def display(self):
        """
        Display all widgets, usually in an ipython notebook.
        """
        for widget in self._widgets.values():
            display(widget)

    def add(self, key, value=None, description=None, callback_functions=None,
            options=None, **kwargs):
        """
        Adds a parameter, which can be represented as a widget.
        Key is the name of the widget.
        The widget is added to the modification and a callback function is
        registered which is called upon any change of the widget.

        callback_functions: array (default [])
            array containing functions (callback functions) to be called upon
            change of the parameter.
        """
        # create and register a widget
        widget = self._create_widget(key, value=value, description=description,
                                     options=options)
        self._widgets[key] = widget

        # Register callback functions to be called upon widget value change
        self._callback_functions[key] = callback_functions or []

        # replace callback_functions by picklable versions
        for i, function in enumerate(self._callback_functions[key]):
            self._callback_functions[key][i] = boundmethod(function)

        # inform ZODB of change
        self._p_changed = True

    def _create_widget(self, key, value=None, description=None, options=None):
        # create widget according to the type of value
        try:
            if value is None:
                value_type = None
            else:
                value_type = type(value)
            if isinstance(value, collections.Iterable) \
                and not isinstance(options, collections.Iterable):
                    options = value
            widget_class = WIDGET_CLASS[value_type]
            widget = widget_class(description=description, value=value,
                                  options=options)

            # register _widget_callback() to be called upon change of widget
            # value
            value_changed = self.__make_widget_callback(self._widget_callback,
                                                        key)
            if value is None:
                widget.on_click(value_changed)
            else:
                widget.observe(value_changed, names='value')
        except:
            # Create FakeWidget in case a graphical widget could not be created
            widget = FakeWidget(value, description)

        return widget

    def __make_widget_callback(self, function, key):
        # needed, see keyword "late binding" (here for key) or
        # http://stackoverflow.com/questions/3431676/creating-functions-in-a-loop
        # Called, whenever one of the values of the widgets is changed.
        def widget_value_changed(name, value=None):
            self._p_changed = True  # inform ZODB of change
            if self._widget_callbacks_active:
                function(key)
        return widget_value_changed

    def _widget_callback(self, key):
        """
        Set a param with key key in self._values to value value.
        call_triggers can be set to False to prevent subsequent calls of
        """
        self._call_cb_functions(key)

    def _call_cb_functions(self, key, **kwargs):
        # call callback_functions
        for function in self._callback_functions[key]:
            if inspect.getargspec(function.mt)[2] is None:
                function()
            else:
                function(**kwargs)

    def set_value(self, key, value, callback=True, **kwargs):
        """
        Sets the value of the widget with key 'key'. Callback functions
        (callback_functions) registered to the widget can be disabled by
        setting callback to False.
        """
        if key in self and value != self._widgets[key].value:
            # disable automatic widget callback
            self._widget_callbacks_active = False

            # set value of widget
            self._widgets[key].value = value
            self._p_changed = True  # inform ZODB of change
            # Is also called by widget_value_changed() in
            # __make_widget_callback() and could be neglected, because setting
            # the immutable value self._widget_callbacks_active already
            # triggers a self._p_changed = True

            # restore automatic widget callback
            self._widget_callbacks_active = True

            # call _callback_functions (what _widget_callback() would do) if
            # callback is wanted
            if callback:
                self._call_cb_functions(key, **kwargs)

    def __contains__(self, key):
        return key in self._widgets

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __getattr__(self, name):
        """
        Allow attributes to be used as params selections for set/get_params
        """
        if name in self:
            return self._widgets[name].value
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        """
        A __setattr__() hook will also override the Persistent __setattr__()
        hook.  User code must treat it much like __getattribute__().  The
        user-defined code must call _p_setattr() first to handle special
        attributes; _p_setattr() takes the attribute name and value.  If it
        returns True, Persistent handled the attribute. If not, the user code
        can run. If the user code modifies the object’s state, it must assigned
        to _p_changed
        """
        if not self._p_setattr(name, value):
            if '_widgets' in self.__dict__ and name in self:
                self.set_value(name, value)
                # is done by set_value:
                # self._p_changed = True # inform ZODB of change
            else:
                super().__setattr__(name, value)

    def __getstate__(self):
        """Return state values to be pickled."""
        # To circumvent: TypeError: can't pickle instancemethod objects
        # __getstate__ should replace instances in self._widgets with
        # information necessary to recreate widgets:
        # widget_classe, description, and value
        state = self.__dict__.copy()
        state['_widgets'] = state['_widgets'].copy()
        for key, widget in state['_widgets'].items():
            value = None
            if hasattr(widget, 'value'):
                value = widget.value
            description = widget.description
            options = None
            if hasattr(widget, 'options'):
                options = widget.options
            # since 0.6.3 options are stored. For compatibility reasons and
            # future optional parameters, additionally store None as a 4th one
            state['_widgets'][key] = [value, description, options, None]
        return state

    def __setstate__(self, state):
        """Restore state from the unpickled state values."""
        # __setstate__ should refill self._widgets with the information stored
        # in self._widgets. Compare with self.add() for necessary steps
        for key, widgetState in state['_widgets'].items():
            if len(widgetState) == 2:
                value, description = widgetState
                options = None
            elif len(widgetState) == 3:  # pre 0.5.0
                widget_class, description, value = widgetState
                if widget_class is widgets.Button:
                    value = None
                options = None
            else:  # len == 4, since 0.6.3
                value, description, options, _ = widgetState
            widget = self._create_widget(key, value=value, options=options,
                                         description=description)
            state['_widgets'][key] = widget
        self.__dict__.update(state)



# -*- coding: utf-8 -*-
"""
Created on Mon Oct 17 18:37:33 2016

@author: Tobias Jachowski
"""
import collections
import os

from . import helpers as hp
from . import version
from .picklable import Attributes


def comp(version):
    return tuple(map(int, (version.split('.'))))


def update_db(experiment):
    """
    Update the format of the file the Experiment is saved in, if the structure
    of the database has changed.
    """
    if not experiment.is_open:
        print("Experiment is closed!")
        return

    old_version = experiment.version
    new_version = version()

    _update_db(experiment, old_version, new_version)


def _update_db(experiment, old_version, new_version):
    if comp(old_version) < comp(new_version):
        print("Updating database from %s to %s ..." % (old_version,
                                                       new_version))

        old_version = comp(old_version)

        # Load analyzer module from tethered_bead package (<= 0.2.3)
        if old_version <= comp('0.2.3'):
            from tethered_bead import analyzer
            # prevent modifications from trying to load data upon loading
            analyzer.Modification.__setstate__ \
                = analyzer.GraphMember.__setstate__

        # Load investigator package (<= 0.5.2)
        if old_version <= comp('0.5.2'):
            import investigator
            # prevent modifications from trying to load data upon loading
            investigator.modification.Modification.__setstate__ \
                = investigator.graph.GraphMember.__setstate__

            # Update all objects to 0.5.1+ (0.5.2), where the module name
            # was changed from investigator to pyoti. If you import the
            # package from the new path only and try to open a database
            # stored with the old package name, you will get broken
            # objects.
            # This means, you have to import the package from the old AND
            # the new path at the same time and then convert all instances
            # that have their classes defined in the investigator package:
            import importlib

            # investigator and pyoti need to be imported ahead
            import pyoti

            def change_object_module(obj, old='investigator', new='pyoti',
                                     new_class_name=None):
                class_name = obj.__class__.__name__
                old_module_name = obj.__module__
                new_module_name = old_module_name.replace(old, new, 1)
                new_module = importlib.import_module(new_module_name)
                if new_class_name is not None:
                    class_name = new_class_name
                obj.__class__ = getattr(new_module, class_name)
                if hasattr(obj, '_p_changed'):
                    obj._p_changed = True

        # Update to 0.2.2
        if old_version < comp('0.2.2'):
            # Update to 0.2.3
            for record in experiment._graphroot.children:
                calibsource = record.calibration.calibsource
                record._traces = ['psdX', 'psdY', 'psdZ',
                                  'positionX', 'positionY', 'positionZ',
                                  'mirrorX', 'laser']
                if hasattr(calibsource, 'C'):
                    calibsource.beta = calibsource.C
                    del calibsource.C
                if hasattr(calibsource, 'mC'):
                    calibsource.mbeta = calibsource.mC
                    del calibsource.mC
                if hasattr(calibsource, 'k'):
                    calibsource.kappa = calibsource.k
                    del calibsource.k
                if hasattr(calibsource, 'mk'):
                    calibsource.mkappa = calibsource.mk
                    del calibsource.mk
                if hasattr(calibsource, 'hsurf'):
                    calibsource.dsurf = calibsource.hsurf
                    del calibsource.hsurf
                # if hasattr(record, '_inversion'):
                    # positionZ was inverted, it now is proportional to the
                    # distance of the bead to the surface and not the
                    # position of the stage anymore.
                    # record._inversion[5] = -1
            for modification in experiment._graphroot.members(
                    instance_class=analyzer.Impact):
                if 'stageZ' in modification._traces_apply:
                    traces = modification._traces_apply
                    i = traces.index('stageZ')
                    traces.remove('stageZ')
                    traces.insert(i, 'positionZ')
                if hasattr(modification, 'wa'):
                    del modification.wa
                if hasattr(modification, 'leftpull'):
                    del modification.leftpull
                if hasattr(modification, 'rightpull'):
                    del modification.rightpull
            for modification in experiment._graphroot.members(
                    instance_class=analyzer.Rotation):
                if hasattr(modification, 'wa'):
                    del modification.wa
                if hasattr(modification, 'stageZ'):
                    del modification.stageZ
            for modification in experiment._graphroot.members(
                    instance_class=analyzer.Attachment):
                if hasattr(modification, 'wa'):
                    del modification.wa
                if hasattr(modification, 'rightstage'):
                    del modification.rightstage
                if hasattr(modification, 'leftstage'):
                    del modification.leftstage
                if hasattr(modification, 'pullleft'):
                    del modification.pullleft
                if hasattr(modification, 'pullright'):
                    del modification.pullright

        # Update to 0.1.3
        if old_version < comp('0.1.3'):
            # Update to version 0.1.3
            for record in experiment._graphroot.children:
                if hasattr(record, 'datasource') \
                        and not hasattr(record.datasource, '_datadir_orig'):
                    try:
                        if hasattr(record.datasource, 'datadir'):
                            datasource = record.datasource
                            datasource._datadir_orig \
                                = record.datasource.datadir
                        if isinstance(record.datasource,
                                      analyzer.ASWADBinData):
                            datasource.name = (
                                "ASWAD LabVIEW bin data originally from \n"
                                "    %s%s%s with \n"
                                "    samplingrate %s Hz") \
                                % (datasource.datadir_orig,
                                   os.sep,
                                   datasource.datafile,
                                   datasource.samplingrate)
                    except:
                        pass

        # Update to 0.1.4
        if old_version < comp('0.1.4'):
            # Update to version 0.2.0
            for modification in experiment.modifications():
                if isinstance(modification, analyzer.Rotation):
                    del modification.stageZ
                    del modification.calibration

        # Update to 0.2.3 database format
        if old_version <= comp('0.2.3'):
            # Retrieve all objects in database ...
            change_object_module(experiment._graphroot,
                                 old='tethered_bead.analyzer',
                                 new='pyoti.graph')
            for obj in experiment._graphroot.members():
                if isinstance(obj, analyzer.Record):
                    # Record
                    change_object_module(obj,
                                         old='tethered_bead.analyzer',
                                         new='pyoti.region.record')
                    # Node of calibration
                    change_object_module(obj.calibration._node,
                                         old='tethered_bead.analyzer',
                                         new='pyoti.graph')
                    # Calibration
                    change_object_module(obj.calibration,
                                         old='tethered_bead.analyzer',
                                         new='pyoti.calibration.calibration')
                    # CalibSource
                    calibsource = obj.calibration.calibsource
                    if isinstance(calibsource, analyzer.CalibrationMATLAB):
                        calibsource.filename = calibsource.calibfile
                        del calibsource.calibfile
                        change_object_module(
                            calibsource,
                            old='tethered_bead.analyzer',
                            new='pyoti.calibration.calibsources.cnmatlab',
                            new_class_name='CNMSource')
                    elif isinstance(calibsource, analyzer.CalibrationTA):
                        calibsource.filename = calibsource.tacfile
                        del calibsource.tacfile
                        change_object_module(
                            calibsource,
                            old='tethered_bead.analyzer',
                            new='pyoti.calibration.calibsources.cnmatlab',
                            new_class_name='CNMSource')
                    else:
                        change_object_module(
                            calibsource,
                            old='tethered_bead.analyzer',
                            new='pyoti.calibration.calibsource')
                    # DataSource
                    datasource = obj.datasource
                    if isinstance(datasource, analyzer.ASWADBinData):
                        change_object_module(
                            datasource,
                            old='tethered_bead.analyzer',
                            new='pyoti.data.datasources.cnlabview',
                            new_class_name='CNLabViewBinData')
                        # Convert ASWADBinData to CNLabViewBinData
                        filename, absdir, fullfilename \
                            = hp.file_and_dir(datasource.datafile,
                                              experiment._graphroot._v_absdir)
                        datasource._filename = filename
                        datasource._directory \
                            = os.path.relpath(absdir,
                                              experiment._graphroot._v_absdir)
                        datasource._filename_orig = filename
                        datasource._directory_orig \
                            = datasource._datadir_orig
                        _fn, datasource._absdir_orig, _ffn \
                            = hp.file_and_dir(datasource.datafile,
                                              datasource._datadir_orig)
                        del datasource.datadir
                        del datasource.datafile
                        del datasource._datadir_orig
                # Node
                change_object_module(obj._node,
                                     old='tethered_bead.analyzer',
                                     new='pyoti.graph')
                # MultiRegion
                if isinstance(obj, analyzer.MultiRegion):
                    change_object_module(obj,
                                         old='tethered_bead.analyzer',
                                         new='pyoti.region.view')
                # View
                if isinstance(obj, analyzer.View):
                    change_object_module(obj,
                                         old='tethered_bead.analyzer',
                                         new='pyoti.region.view')
                # InteractiveAttributes of Modification
                if isinstance(obj, analyzer.Modification):
                    change_object_module(obj.iattributes,
                                         old='tethered_bead.analyzer',
                                         new='pyoti.picklable')
                # Offset modifiation
                if isinstance(obj, analyzer.Offset):
                    change_object_module(
                        obj,
                        old='tethered_bead.analyzer',
                        new='pyoti.modification.modifications.offset')
                # Impact modification
                if isinstance(obj, analyzer.Impact):
                    change_object_module(
                        obj,
                        old='tethered_bead.analyzer',
                        new='pyoti.modification.modifications.impact')
                # Beadscan modification
                if isinstance(obj, analyzer.Beadscan):
                    change_object_module(
                        obj,
                        old='tethered_bead.analyzer',
                        new='pyoti.modification.modifications.beadscan')
                # Attachment modification
                if isinstance(obj, analyzer.Attachment):
                    change_object_module(
                        obj,
                        old='tethered_bead.analyzer',
                        new='pyoti.modification.modifications.attachment')
                # Rotation modification
                if isinstance(obj, analyzer.Rotation):
                    change_object_module(
                        obj,
                        old='tethered_bead.analyzer',
                        new='pyoti.modification.modifications.rotation')
                # Modification boundmethod and evaluate button
                if isinstance(obj,
                              pyoti.modification.modification.Modification):
                    # picklable_boundmethod -> pyoti.picklable.boundmethod
                    for functions in \
                            obj.iattributes._callback_functions.values():
                        for fun in functions:
                            change_object_module(fun,
                                                 old='tethered_bead.analyzer',
                                                 new='pyoti.picklable',
                                                 new_class_name='boundmethod')
                    # Add evaluate button
                    obj.add_iattribute('evaluate', description='Evaluate',
                                       unset_automatic=False,
                                       set_changed=False,
                                       callback_functions=[obj.evaluate])

        # Switch off inversion of 'positionZ' before version 0.2.3
        if old_version <= comp('0.2.3'):
            for record in experiment._graphroot.children:
                record.set_inversion('positionZ', False)

        # Update to 0.5.0
        if old_version < comp('0.5.0'):
            # The database structure for iattributes has changed.
            # Modifications are the only objects using iattributes.
            # Update all iattributes by triggering their unpickling and
            # informing ZODB of change
            # The update itself is done in the method:
            # picklable.InteractiveAttributes.__setstate__()
            for modification in experiment.modifications():
                modification.iattributes._p_changed = True

        # Update to 0.5.2 database format
        if old_version <= comp('0.5.2'):
            if experiment.cached_region is None:
                experiment.cached_region = []
            if not isinstance(experiment.cached_region, collections.Iterable):
                experiment.cached_region = [experiment.cached_region]

            # Retrieve all objects in database ...
            for obj in experiment._graphroot.members():
                if isinstance(obj, pyoti.region.Record) \
                        or isinstance(obj, investigator.region.Record):
                    # Node of calibration
                    change_object_module(obj.calibration._node)
                    # Calibration
                    change_object_module(obj.calibration)
                    # CalibSource
                    change_object_module(obj.calibration.calibsource)
                    # DataSource
                    change_object_module(obj.datasource)
                if isinstance(obj, pyoti.modification.Modification) \
                    or isinstance(obj,
                                  investigator.modification.Modification):
                    change_object_module(obj.iattributes)
                # Node
                change_object_module(obj._node)
                # Cargo (Record, MultiRegion, View, Modification)
                change_object_module(obj)

            # change_object_module(experiment._dbroot['status'])
            # experiment._dbroot._p_changed = True
            new_status = Attributes()
            old_status = experiment._dbroot['status']
            new_status.cached_region = old_status.cached_region
            if hasattr(old_status, 'last_added_region'):
                new_status.last_added_region = old_status.last_added_region
            else:
                new_status.last_added_region = None
            if hasattr(old_status, 'last_adjusted_view'):
                new_status.last_adjusted_view = old_status.last_adjusted_view
            else:
                new_status.last_adjusted_view = None
            experiment._dbroot['status'] = new_status

        # Update to 0.5.3
        if old_version < comp('0.5.3'):
            for child in experiment._graphroot.children:
                if (hasattr(child, 'datasource')
                        and hasattr(child.datasource, 'directory')):
                    try:
                        child.datasource._load_file = True
                        child.datasource._root = experiment._graphroot
                        child.datasource._filename \
                            = child.datasource._filename_orig
                        child.datasource._directory = '.'
                        # Unfortunately, we can't convert the relative
                        # directory_orig to an absolute path, because we don't
                        # know, whether the file is opened from the same
                        # location, during the update.
                        child.datasource._absdir_orig \
                            = child.datasource._directory_orig
                        del child.datasource.__dict__['directory']
                        del child.datasource.__dict__['filename']
                        # The last one could fail, but this doesn't matter.
                        del child.datasource.__dict__['parafile']
                    except:
                        pass

        # Update to 0.5.4
        if old_version < comp('0.5.4'):
            for region in experiment.regions():
                if hasattr(region, '_data_cached'):
                    del region.__dict__['_data_cached']
                    region._p_changed = True

        # Update to 0.6.0
        if old_version < comp('0.6.0'):
            for view in experiment.views():
                sticky = True
                if hasattr(view, 'autonomous'):
                    sticky = not view.autonomous
                    del view.__dict__['autonomous']
                view.sticky_start = sticky
                view.sticky_stop = sticky

        # Update to 0.6.1
        if old_version < comp('0.6.1'):
            for mod in experiment.modifications():
                try:
                    if hasattr(mod, 'ifigure'):
                        del mod.__dict__['ifigure']
                        mod._p_changed = True
                except:
                    pass
                try:
                    if hasattr(mod, 'rfigure'):
                        del mod.__dict__['rfigure']
                        mod._p_changed = True
                except:
                    pass

        # Update to 0.6.2
        if old_version < comp('0.6.2'):
            for child in experiment._graphroot.children:
                if hasattr(child, 'datasource'):
                    ds = child.datasource
                    # force datasource to be unpickled
                    if hasattr(ds, '_load_file') \
                            and '_load_file' in ds.__dict__:
                        ds._needs_file = ds.__dict__['_load_file']
                        del ds.__dict__['_load_file']

        # Update to 0.8.0
        if old_version < comp('0.8.0'):
            # Convert all GModifications to corresponding new Modifications
            # GAttachment -> Attachment
            # GBaseline-> Baseline
            # GBeadscan -> Beadscan
            #  Beadscan.spline -> Beadscan.model
            #  delete beadscan.bin_means
            # GImpact -> Touchdown
            #  pyoti.modification.modifications.impact
            #      -> pyoti.modification.modifications.touchdown
            #  Impact -> Touchdown
            #  iattributes.impact -> iattributes.touchdown
            #  iattributes.set_impact_calib -> iattributes.set_touchdown_calib
            #   delete impact.bin_means
            # GRotation -> Rotation
            # for all iattributes datapoints_used -> datapoints
            from .modification import \
                Attachment, Baseline, Beadscan, Touchdown, Rotation

            MODIFICATION_CLASS = {
                'GAttachment': Attachment,
                'GBaseline': Baseline,
                'GBeadscan': Beadscan,
                'GImpact': Touchdown,
                'GRotation': Rotation
            }

            def create_new_mod(old_mod, new_modclass):
                view_based = old_mod.view_based
                view_apply = old_mod.view_apply
                name = old_mod.name
                new_mod = new_modclass(view_based=view_based,
                                       view_apply=view_apply,
                                       name=name, group=name,
                                       db_update=True)
                new_mod._traces_apply = old_mod.traces_apply
                if hasattr(old_mod.iattributes, 'automatic'):
                    new_mod.iattributes.automatic \
                        = old_mod.iattributes.automatic
                return new_mod

            def set_iattribute_value(old_mod, new_mod, key, new_key=None):
                new_key = new_key or key
                value = old_mod.iattributes[key]
                new_mod.iattributes.set_value(new_key, value, callback=False)

            # create new mods
            for old_mod in experiment.modifications():
                old_class_name = old_mod.__class__.__name__
                try:
                    new_class = MODIFICATION_CLASS[old_class_name]
                except:
                    continue
                new_mod = create_new_mod(old_mod, new_class)

                # transfer individual attributes
                if old_class_name == 'GBaseline':
                    new_mod._baseline_idx = old_mod._baseline_idx.copy()
                    new_mod._model = old_mod.model.copy()
                if old_class_name == 'GBeadscan':
                    new_mod._model = old_mod.spline.copy()
                    new_mod.minpsdZ = old_mod.minpsdZ
                    new_mod.maxpsdZ = old_mod.maxpsdZ
                if old_class_name == 'GImpact':
                    new_mod.left_upper = old_mod.left_upper
                    new_mod.right_upper = old_mod.right_upper
                    new_mod.left = old_mod.left
                    new_mod.right = old_mod.right
                if old_class_name == 'GRotation':
                    new_mod.rotation_method = old_mod.rotation_method

                # transfer iattributes values
                for key in old_mod.iattributes._widgets:
                    new_key = key
                    if key == 'datapoints_used':
                        new_key = 'datapoints'
                    if key == 'impact':
                        new_key = 'touchdown'
                    if key == 'set_impact_calib':
                        new_key = 'set_touchdown_calib'
                    set_iattribute_value(old_mod, new_mod, key, new_key)

                # remove old mod
                old_mod.view_based.remove_child(old_mod)
                # Second, delete relation to child
                old_mod.view_apply.remove_parent(old_mod)

        # Update to 0.9.0
        if old_version < comp('0.9.0'):
            # remove '_mod' from the group attribute. Since 0.8.1 name and
            # group can be independently set.
            for member in experiment.members():
                member.group = member.group.replace('_mod', '')
                if hasattr(member, 'iattributes'):
                    widget = member.iattributes._widgets['active']
                    widget.description = widget.description.replace(
                        "modification ", "")
                    if "datapoints" in member.iattributes._widgets:
                        widget = member.iattributes._widgets['datapoints']
                        widget.description = widget.description.replace(
                            "used ", "")

        # Update to 0.11.0
        if old_version < comp('0.11.0'):
            # Map old modules to new module names (new modules have already
            # been loaded by plugin.plugin_loader.load_modules())
            import sys
            import pyoti
            old_mods = ['pyoti.modification.modifications.attachment',
                        'pyoti.modification.modifications.baseline',
                        'pyoti.modification.modifications.beadscan',
                        'pyoti.modification.modifications.offset',
                        'pyoti.modification.modifications.rotation',
                        'pyoti.modification.modifications.touchdown',
                        'pyoti.modification.modifications.generic',
                        'pyoti.data.datasources.cnlabview',
                        'pyoti.data.datasources.generic',
                        'pyoti.calibration.calibsources.cellnano',
                        'pyoti.calibration.calibsources.pyotic']
            new_mods = ['pyoti.plugins.modifications.attachment',
                        'pyoti.plugins.modifications.baseline',
                        'pyoti.plugins.modifications.beadscan',
                        'pyoti.plugins.modifications.offset',
                        'pyoti.plugins.modifications.rotation',
                        'pyoti.plugins.modifications.touchdown',
                        'pyoti.plugins.modifications.generic',
                        'pyoti.plugins.datasources.cellnano',
                        'pyoti.plugins.datasources.generic',
                        'pyoti.plugins.calibsources.cellnano',
                        'pyoti.plugins.calibsources.pyotic']
            for old, new in zip(old_mods, new_mods):
                sys.modules[old] = sys.modules[new]

        # Update to 0.10.1
        if old_version < comp('0.10.1'):
            # Rename CNMSource to CNMatlabSource
            # pyoti.calibration.calibsources.cnmatlab.CNMSource ->
            # pyoti.calibration.calibsources.cellnano.CNMatlabSource
            for record in experiment.records():
                old_calibsource = record.calibration.calibsource
                old_class_name = old_calibsource.__class__.__name__
                if old_class_name == 'CNMSource':
                    change_object_module(
                        old_calibsource,
                        old='pyoti.calibration.calibsources.cnmatlab',
                        new='pyoti.calibration.calibsources.cellnano',
                        new_class_name='CNMatlabSource')

        # Update to 0.11.0
        if old_version < comp('0.11.0'):
            # To update the __module__ attribute of objects, one must
            # trigger a change of objects directly referencing the objects,
            # whose __module__ has changed, not the objects themselves. This is
            # weird!
            for obj in experiment._graphroot.members():
                if isinstance(obj, pyoti.modification.Modification):
                    # Update modification (directly referenced by a node)
                    obj._node._p_changed = True
                if isinstance(obj, pyoti.region.Record):
                    # Update datasource (directly referenced by a record)
                    obj._p_changed = True
                    # Update calibsource (directly referency by a calibration)
                    obj.calibration._p_changed = True

        # Update to 0.14.0
        # change to SI units
        if old_version < comp('0.14.0'):
            import pyoti
            for record in experiment.records():
                record._conversion[record.traces_to_idx('positionXYZ')] *= 1e-6
                if 'mirrorX' in record.traces:
                    record._conversion[record.traces_to_idx('mirrorX')] *= 1e-6
                if 'mirrorY' in record.traces:
                    record._conversion[record.traces_to_idx('mirrorY')] *= 1e-6
                record._p_changed = True
                calibsource = record.calibration.calibsource
                calibsource.beta *= 1e-6  # nm/mV -> m/V
                #calibsource.mbeta *= 1  # nm/mV/µm -> m/V/m
                calibsource.kappa *= 1e-3  # pN/nm -> N/m
                calibsource.mkappa *= 1e3  # pN/nm/µm -> N/m/m
                calibsource.dsurf *= 1e-6  # µm -> m
                calibsource.radiusspec *= 1e-6  # µm -> m
            for mod in experiment.modifications():
                class_name \
                    = ''.join((mod.__module__, '.', mod.__class__.__name__))
                if class_name \
                        == 'pyoti.plugins.modifications.touchdown.Touchdown':
                    mod.iattributes.touchdown *= 1e-6
                    mod.iattributes._widgets['touchdown'].description = 'Touchdown (m)'
                if class_name \
                        == 'pyoti.plugins.modifications.attachment.Attachment':
                    mod.iattributes.offsetStage *= 1e-6
                    mod.iattributes._widgets['offsetStage'].description = 'Offset position (m)'
                if class_name \
                        == 'pyoti.plugins.modifications.rotation.Rotation':
                    # Update rotation method from nm to m
                    mod.rotation_method = 'm'


        # Update to X.Y.Z
        # if old_version < comp('X.Y.Z'):
            # for record in experiment.records():
            # for modification in experiment._graphroot.members(
            #                                           instance_class=Impact):
            # for record in experiment._graphroot.children:
            # for modification in experiment.modifications():
            # experiment._dbroot['version'] = 'X.Y.Z'

        # If there were no database layout changes, automatically set db
        # version to the newest version
        if comp(new_version) > comp(experiment._dbroot['version']):
            experiment._dbroot['version'] = new_version



# -*- coding: utf-8 -*-
"""
Created on Thu Mar 31 14:58:54 2016

@author: Tobias Jachowski
"""
import numpy as np

from pyoti.modification.modification import Modification
from pyoti import helpers as hp
from pyoti import traces as tc
from pyoti.picklable import unboundfunction


class GenericMod(Modification):
    """
    Generic modification that modifies a set of chosen traces.
    """
    def __init__(self, modify=None, recalculate=None, mod_params=None,
                 print_info=None, **kwargs):
        # modify: function, that modifies the data
        # recalculate: function, which (in automatic mode) is called before
        #               _modify(). Can be used to automatically calculate
        #               mod_params, which are calculated with data obtained
        #               from view_based, i.e. data obtained by
        #               _get_data_based(). (Only called, if self.updated
        #               is False. Self.updated is set to False, upon any change
        #               of the view_based). The mod_params can be used by
        #               _modify().
        # kwargs:
        # traces_apply=None,
        # view_apply=None, # obligatory
        # view_based=None,
        # automatic_switch=False, Option to disable automatic call of
        #           _recalculate() before the call of _modify()
        # datapoints=-1, if>0 creates an iattribute, which is used for
        #                     decimate/average calculation in functions
        #                   _get_data(), _get_data_apply, and _get_data_based()
        #                  and bins determination in calculate_bin_means()

        if modify is None:
            raise TypeError("GenericMod missing required positional argument "
                            "`modify`.")

        # initialize _mod_params before call of super().__init__
        # to make sure __getattr__() doesn't end in a infinite
        # recursion loop
        self._mod_params = []

        super().__init__(**kwargs)

        # register widgets for modifications of traces
        for trace in self.traces_apply:
            self._mod_params.append(trace)
            key = self._key(trace)
            description = ''.join(('Parameter for trace ', tc.label(trace)))
            self.add_iattribute(key, description=description, value=0.0)

        if mod_params is not None:
            mod_params = hp.listify(mod_params)
            for name in mod_params:
                self._mod_params.append(name)
                key = self._key(name)
                description = ''.join(('Parameter ', name))
                self.add_iattribute(key, description=description, value=0.0)

        self._mod = unboundfunction(modify)

        if recalculate is not None:
            self._recalc = unboundfunction(recalculate)

        if print_info is not None:
            self._print_inf = unboundfunction(print_info)

    def _modify(self, data, samples, data_traces, data_index, mod_index):
        return self._mod(self, data, samples, data_traces, data_index,
                         mod_index)

    def _recalculate(self):
        if hasattr(self, '_recalc'):
            self._recalc(self)

    def _print_info(self):
        if hasattr(self, '_print_inf'):
            self._print_inf()

    def _key(self, trace):
        return ''.join(('mod_param_', trace))

    def get_mod_params(self, names=None):
        if names is None:
            # trace should have the same dimension as traces_apply
            names = self._mod_params
        else:
            names = hp.listify(names)

        return np.array([self.iattributes[self._key(name)] for name in names])

    def set_mod_params(self, values, names=None, leave_automatic=False):
        if names is None:
            # trace should have the same dimension as mod_params
            names = self._mod_params
        else:
            names = hp.listify(names)
        values = hp.listify(values)

        for name, value in zip(names, values):
            key = self._key(name)
            self.iattributes.set_value(key, value,
                                       leave_automatic=leave_automatic)

    def __getattr__(self, name):
        """
        Allow attributes to be used as params selections for set/get_params
        """
        if name in self._mod_params:
            key = self._key(name)
        # if key in self.iattributes:
            return self.iattributes[key]
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        # key = self._key(name)
        if not self._p_setattr(name, value):
            if '_mod_params' in self.__dict__ and name in self._mod_params:
                self.set_mod_params(value, name)
            else:
                super().__setattr__(name, value)

    mod_params = property(get_mod_params, set_mod_params)



# -*- coding: utf-8 -*-
"""
Created on Wed Mar  9 17:09:09 2016

@author: Tobias Jachowski
"""
import collections
import numpy as np
try:
    import pandas as pd
    __pd__ = True
except ImportError:
    __pd__ = False
from abc import ABCMeta, abstractmethod, abstractproperty
from scipy.ndimage import convolve1d
from scipy.ndimage import median_filter

from .. import helpers as hp
from .. import traces as tc
from ..graph import GraphMember
from ..evaluate import signal as sn


class Region(GraphMember, metaclass=ABCMeta):
    """
    A Region provides common methods, to define and reference regions on given
    data, and methods to access the data in the defined regions.
    A Region of the data to refer to can be choosen by setter methods
    (start/stop or tmin/tmax). The region can be defined as a timespan
    (seconds) or an indexspan (indexes of a numpy array).
    """

    def __init__(self, caching=False, **kwargs):
        """
        Parameters
        ----------
        caching : bool
        **kwargs

        Attributes
        ----------
        start
        stop
        tmin
        tmax
        timespan
        samplingrate
        datapoints
        seconds
        indexspan
        indexvector
        timevector
        num_traces
        caching
        traces
        data
        """
        super().__init__(**kwargs)
        self.caching = caching

    @abstractproperty
    def start(self):
        pass

    @abstractproperty
    def stop(self):
        pass

    @start.setter
    def start(self, start):
        pass

    @stop.setter
    def stop(self, stop):
        pass

    @property
    def tmin(self):
        return self.start / self.samplingrate

    @property
    def tmax(self):
        return (self.stop - 1) / self.samplingrate

    @tmin.setter
    def tmin(self, tmin):
        self.start = int(np.around(tmin * self.samplingrate))

    @tmax.setter
    def tmax(self, tmax):
        self.stop = int(np.around(tmax * self.samplingrate)) + 1

    @property
    def timespan(self):
        return (self.tmin, self.tmax)

    @timespan.setter
    def timespan(self, timespan):
        self.tmin, self.tmax = timespan

    def set_timespan(self, tmin, tmax):
        self.tmin = tmin
        self.tmax = tmax

    @abstractproperty
    def samplingrate(self):
        # points per second
        pass

    @property
    def resolution(self):
        return self.samplingrate

    @property
    def datapoints(self):
        return self.stop - self.start

    @property
    def seconds(self):
        return self.datapoints / self.samplingrate

    @property
    def indexspan(self):
        # return slice(self.start, self.stop)
        return slice(0, self.datapoints, 1)

    @property
    def indexvector(self):
        # return np.arange(self.start, self.stop)
        return np.arange(self.datapoints)

    @property
    def timevector(self):
        return self.indexvector / self.samplingrate

    @abstractproperty
    def num_traces(self):
        return self.data.shape[0]

    @property
    def caching(self):
        return self._caching

    @caching.setter
    def caching(self, caching):
        """
        If memory consumption is no concern, use caching to speed up data
        retrieval.
        """
        self._caching = caching
        if not caching:
            # delete cache to free memory, ZODB volatile
            self._v_data_cached = None

    def get_data(self, traces=None, samples=None, moving_filter='mean',
                 window=1, decimate=1, copy=True, time=False, pandas=False):
        """
        Returns data of this region.

        Parameters
        ----------
        traces : str or list of str, optional
        samples : int, list of int, or slice, optional
            Samples start/stop has to be in between 0 and self.datapoints. If
            start, stop or any other index is negative it will be converted to
            a positive index, referring to self.datapoints.
        moving_filter : str, optional
            The moving filter to apply to the data. Can be one of 'mean' or
            'median'. Only applied, if `window` > 1.
        window : int, optional
            The window size of the moving filter.
        decimate : int, optional
            Decimate the data by given value. If `samples` is of type slice and
            `samples.step` is explicitly set (i.e. not None), the step value
            takes precedence over `decimate`.
        copy : bool, optional
            Only if `self.caching` is True, `copy` has the effect of either
            returning a copy of the cached data (`copy`=True) or returning a
            reference to the cached data array.
            If `self.caching` is False, return the uncached data. Then, `copy`
            has no effect at all.
        time : bool, optional
            Add the time as the first trace.
        pandas : bool, optional
            Return a pandas.DataFrame with proper indices and columns.

        Returns
        -------
        2D numpy.ndarray or pandas.DataFrame
            If self.caching is True, return cached data.
        """
        # Set the traces and the samples to proper default values and/or
        # convert (normalize) them to proper formats
        traces_idx = self.traces_to_idx(traces)
        samples = self.samples_idx(samples, decimate)

        if time:
            timevector = self.timevector[samples]

        if window <= 1:
            # Return unfiltered data
            data = self._get_data(samples, traces_idx, copy)
        else:
            # Moving filter should be applied, filter first, then decimate.
            # Get data, disregarding decimating factor, but considering traces
            if isinstance(samples, slice):
                samples_start = samples.start
                samples_stop = samples.stop
            else:  # samples is numpy.ndarray
                samples_start = samples.min()
                samples_stop = samples.max()
            # Get requested and preceding and succeeding data, to avoid
            # boundary issues of the filter
            filter_start = max(0, samples_start - window)
            filter_stop = min(samples_stop + window, self.datapoints)
            filter_samples = slice(filter_start, filter_stop, 1)
            data = self._get_data(filter_samples, traces_idx, copy=False)

            # Filter the data
            data = _moving_filter(data, window, moving_filter=moving_filter)

            # Return requested samples, considering decimating factor
            if isinstance(samples, slice):
                start = samples_start - filter_start
                stop = samples_stop - filter_start
                samples = slice(start, stop, samples.step)
            else:  # samples is numpy.ndarray
                samples = samples - filter_start
            data = data[samples]

        if pandas and __pd__:
            return pd.DataFrame(data, index=self.timevector[samples],
                                columns=self.idx_to_traces(traces_idx))

        if time:
            return np.c_[timevector, data]

        return data

    def _get_data(self, samples, traces_idx, copy=True):
        """
        Returns the data for the given samples and traces_idx.
        """
        # Retrieve cached data
        if self.caching:
            # If the cache is outdated or not yet created, do so
            self.update_cache()

            # Check, if traces and/or samples are instances of numpy.ndarray
            traces_is_ndarray = isinstance(traces_idx, np.ndarray)
            samples_is_ndarray = isinstance(samples, np.ndarray)

            # If at least one of the indices is an ndarray, advanced indexing
            # is triggered.
            if traces_is_ndarray or samples_is_ndarray:
                # If both, traces and samples are an ndarray, indexing two
                # dimensions in a single step would result in an error in
                # numpy. Therefore, separate the indexing into two steps.
                if traces_is_ndarray and samples_is_ndarray:
                    # First, select the samples.
                    data_ = self._v_data_cached[samples]

                    # Second, select the traces.
                    return data_[:, traces_idx]

                # Advanced indexing automatically returns a copy of the data.
                # There is no need to extra copy via ndarray.copy(), which
                # would cost some extra execution time.
                copy = False

            if copy:
                # Per default return a copy of the data cache to protect it
                # from unwanted in place modifications
                return self._v_data_cached[samples, traces_idx].copy()
            else:
                # Only, if explicitly asked for "copy=False", return a
                # direct reference to the cached data
                return self._v_data_cached[samples, traces_idx]

        # No caching enabled, return uncached data
        return self._get_data_uncached(samples, traces_idx, copy)

    def update_cache(self, force=False):
        """
        Create and update cached data.

        Parameters
        ----------
        force : bool
            Recalculate, even if self._v_data_cached is up to date.
        """
        # Check for whether an update of the cache is needed or not.
        if self.caching and (not hasattr(self, '_v_data_cached')
                             or self._v_data_cached is None
                             or force):
            # An update is needed, calculate data for self.indexspan and
            # all traces_idx and store it in the cache. ZODB volatile.
            self._v_data_cached = self._get_data_uncached(self.indexspan,
                                                          self.traces_to_idx(),
                                                          copy=True)

    @abstractmethod
    def _get_data_uncached(self, samples, traces_idx, copy=True):
        pass

    def member_changed(self, ancestor=True, calledfromself=False,
                       leave_cache=False, **kwargs):
        # If self triggered a change, delete cache and trigger an update of the
        # cache. `leave_cache` can prevent deleting of the cache.
        if calledfromself and not leave_cache:
            self._v_data_cached = None  # ZODB volatile
        # If an ancestor triggered a change, delete cache and trigger an update
        # of the cache. A triggered change of descendants is ignored.
        if not calledfromself and ancestor:
            self._v_data_cached = None  # ZODB volatile

        # Call method of superclass `GraphMember`
        super().member_changed(ancestor=ancestor,
                               calledfromself=calledfromself, **kwargs)

    def traces_available(self, traces=None):
        traces_idx = self.traces_to_idx(traces)
        return self.idx_to_traces(traces_idx)

    def traces_to_idx(self, traces=None):
        """
        return index/slice of trace/s
        if trace is None, default to all traces
        """
        # Default to all traces
        if traces is None:
            traces_idx = slice(0, self.num_traces, 1)
        # If traces is a slice, correct for negative and None values
        elif isinstance(traces, slice):
            # TODO: if step is negative, indices() evaluates stop to
            # -1, but it should be -1 -self.num_traces!
            start, stop, step = traces.indices(self.num_traces)
            traces_idx = slice(start, stop, step)
        else:
            traces_idx = []
            for trace in tc.normalize(traces):
                if trace in self.traces:
                    traces_idx.append(self.traces.index(trace))
                elif isinstance(trace, int) and trace < self.num_traces:
                    traces_idx.append(trace)

        return hp.slicify(traces_idx, length=self.num_traces)

    def idx_to_traces(self, traces_idx=None):
        if isinstance(traces_idx, slice):
            traces_idx = hp.listify(traces_idx)
        if traces_idx is None:
            traces = self.traces
        elif isinstance(traces_idx, collections.Iterable):
            traces = [self.traces[trace_idx]
                      for trace_idx in traces_idx
                      if trace_idx < self.num_traces]
        return traces

    def samples_idx(self, samples=None, decimate=1):
        """
        Returns a slice or an np.array.

        Parameters
        ----------
        samples : int or list of int or slice, optional
            If samples is None, return the slice spanning the whole Region:
            slice(0, `self.datapoints`, `decimate`).
            If `samples` is an instance of int, slice, list or np.array,
            negative indices (i_neg) are converted into positive ones (i_pos):
            i_pos = i_neg + `self.datapoints`
        decimate : int, optional
            The decimate value of the returned slice, or the value the index
            array is decimated with.
            If `samples` is of type slice and `samples.step` is not explicitly
            set (i.e. None), `decimate` takes precedence over the step value.
        """
        if samples is None:
            samples = slice(0, self.datapoints, decimate or 1)
        elif isinstance(samples, int):
            if samples < 0:
                # correct for negative indices in original samples
                # referring to self.datapoints
                samples += self.datapoints
            samples = slice(samples, samples + 1, decimate or 1)
        elif isinstance(samples, slice):
            # Correct samples for negative indices and None
            start, stop, step = samples.indices(self.datapoints)

            # If samples.step is not explicitly set (i.e. None), decimate takes
            # precedence of the step information (1 per default)
            step = samples.step or decimate or 1

            samples = slice(start, stop, step)
        else:
            # samples either a list or an np.array
            # convert list to array and apply decimate (1 per default)
            if isinstance(samples, list):
                samples = np.array(samples)[::decimate or 1]
            # correct for negative indices in original samples
            # referring to self.datapoints
            samples[samples < 0] += self.datapoints

        return hp.slicify(samples)

    def _excited(self, traces=None, index=True):
        """
        For backward compatibility in pyotc.

        Parameters
        ----------
        traces : list of str
            Two traces to be compared. Defaults to ['positionX', 'positionY'].

        Returns
        -------
        trace : str
            The excited trace.
        """
        traces = traces or ['positionX', 'positionY']
        data = self.get_data(traces=traces, copy=False)
        ex = sn.get_excited_signal(data)
        if index:
            return ex
        return traces[ex]

    @abstractproperty
    def traces(self):
        return

    @property
    def data(self):
        # return all data
        return self.get_data()

    def __getattr__(self, name):
        """
        Allow attributes to be used as trace selections for get_data
        """
        if name in self.traces or name in tc:
            # name is directly known by Region or
            # name is probably an alias, a shorthand notation, or a combination
            return self.get_data(traces=name)
        else:
            raise AttributeError(name)


def _moving_filter(data, window, moving_filter='mean', mode='reflect', cval=0.0,
                   origin=0):
    """
    Apply a moving filter to data.

    Parameters
    ----------
    data : numpy.ndarray
        The data to be filterd.
    window : int
        The window size of the moving filter.
    moving_filter : str, optional
        The filter to be used for the moving filter. Can be one of 'mean' or
        'median'.
    mode : str, optional
        mode       |   Ext   |         Data           |   Ext
        -----------+---------+------------------------+---------
        'mirror'   | 4  3  2 | 1  2  3  4  5  6  7  8 | 7  6  5
        'reflect'  | 3  2  1 | 1  2  3  4  5  6  7  8 | 8  7  6
        'nearest'  | 1  1  1 | 1  2  3  4  5  6  7  8 | 8  8  8
        'constant' | 0  0  0 | 1  2  3  4  5  6  7  8 | 0  0  0
        'wrap'     | 6  7  8 | 1  2  3  4  5  6  7  8 | 1  2  3
        See 'scipy.ndimage.convolve1d' or 'scipy.ndimage.median_filter'
    cval : float, optional
        See 'scipy.ndimage.convolve1d' or 'scipy.ndimage.median_filter'
    origin : int, optional
        See 'scipy.ndimage.convolve1d' or 'scipy.ndimage.median_filter'
    """
    mode = mode or 'reflect'
    cval = cval or 0.0
    origin = origin or 0
    if moving_filter == 'mean' or moving_filter == 'average':
        return _movingmean(data, window, mode=mode, cval=cval, origin=origin)
    else:  # if moving == 'median'
        return _movingmedian(data, window, mode=mode, cval=cval, origin=origin)


def _movingmean(data, window, mode='reflect', cval=0.0, origin=0):
    weights = np.repeat(1.0, window)/window
    # sma = np.zeros((data.shape[0] - window + 1, data.shape[1]))
    sma = convolve1d(data, weights, axis=0, mode=mode, cval=cval,
                     origin=origin)
    return sma


def _movingmedian(data, window, mode='reflect', cval=0.0, origin=0):
    if data.ndim == 1:
        size = window
    else:
        size = (window, 1)
    smm = median_filter(data, size=size, mode=mode, cval=cval, origin=origin)
    return smm


def _moving_mean(data, window):
    """
    Calculate a filtered signal by using a moving mean. The first datapoint is
    the mean of the first `window` datapoints and the last datapoint is the mean
    of the last `window` datapoints of the original data. This function does not
    handle the lost edges of the data, i.e. the filtered data is shortened by
    `window` datapoints.

    This function is faster than the function `movingmean()`.

    Parameters
    ----------
    data : 1D numpy.ndarray of type float
        Data to calculate the rolling mean from.
    window : int
        Length of the window to calculate the rolling mean with.

    Returns
    -------
    1D numpy.ndarray of type float
        The data filtered with a rolling mean.
    """
    cumsum = np.cumsum(np.insert(data, 0, 0))
    return (cumsum[window:] - cumsum[:-window]) / window


def _moving_mean_pandas(data, window):
    """
    Calculate a filtered signal by using a moving mean.

    Parameters
    ----------
    data : 1D numpy.ndarray of type float
        Data to calculate the rolling mean from.
    window : int
        Length of the window to calculate the rolling mean with.

    Returns
    -------
    1D numpy.ndarray of type float
        The data filtered with a rolling mean.
    """
    data = pd.Series(data)
    r = data.rolling(window=window)
    return r.mean()[window - 1:].get_values()



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 12 14:22:31 2016

@author: Tobias Jachowski
"""
import collections
import matplotlib.pyplot as plt
import numpy as np
from abc import ABCMeta, abstractmethod

from .. import gui
from .. import helpers as hp
from .. import traces as tc
from ..evaluate import signal as sn
from ..graph import GraphMember
from ..picklable import InteractiveAttributes


class GraphicalMod(object):
    """
    This class's subclasses should implement `_figure()` and `_update_fig()`,
    which return and update a matplotlib figure, respectively. The figure can
    be accessed by `self.figure`.

    Parameters
    ----------
    figure
    modification : Modification
    """
    def __init__(self, modification=None, **kwargs):
        # Register the modification which should be graphically adjusted
        self.modification = modification

        # Initialize figure to None, which effectively disables
        # `self.update_fig()` and Co. and prevent them from throwing an error
        self._fig = None

    def _set_plot_params(self, plot_params=None):
        if plot_params is None:
            plot_params = {}
        gui.set_plot_params(plot_params=plot_params)

    def display(self, plot_params=None):
        self.init_fig(plot_params=plot_params)

    def init_fig(self, show=True, plot_params=None):
        """
        This method calls self._figure() to create an interactive figure and
        interact with the user to determine the parameters necessary to
        calculate the modification (see self._recalculate()). and
        self._close_fig() to release all references to the actors of the
        figure.
        `self._figure()` and self._close_fig() should be (over)written by
        subclasses.
        """
        # Only create a figure, if the function `self._figure()` is implemented
        if not hasattr(self, '_figure'):
            return

        # close the figure
        # nbagg backend needs to have the figure closed and recreated
        # whenever the code of the cell displaying the figure is executed.
        # A simple update of the figure would let it disappear. Even a
        # self.figure.show() wouldn't work anymore.
        # For backends this just means a bit of extra calculation.
        # Therefore, close the figure first before replotting it.
        self.close_fig()

        # set default plot parameters, can be recalled / overwritten in
        # `self._figure()`
        self._set_plot_params(plot_params=plot_params)

        # create the figure
        self.figure = self._figure()

        # update the figure
        self.update_fig()

        # show the figure
        if show:
            self.figure.show()

    def update(self, **kwargs):
        self.update_fig(**kwargs)

    def update_fig(self, **kwargs):
        if self._fig is not None:
            self._update_fig(**kwargs)
            self._figure_canvas_draw()

    def _update_fig(self, **kwargs):
        pass

    def close_fig(self):
        if self._fig is not None:
            self._pre_close_fig()
            self._close_fig()
            self._post_close_fig()

    def _pre_close_fig(self):
        """
        Method to be overwritten by subclasses.
        """
        pass

    def _close_fig(self):
        # force redraw of the figure
        self._figure_canvas_draw()

        # close the figure
        plt.close(self.figure)

        # release memory
        self.figure = None

    def _post_close_fig(self):
        """
        Method to be overwritten by subclasses.
        """
        pass

    def _figure_canvas_draw(self):
        # Some matplotlib backends will throw an error when trying to draw the
        # canvas. Simply ignoring the error that could happen here will prevent
        # the figure from not beeing closed, left open, and preventing the next
        # figure to be drawn. Even though the "except: pass" clause is
        # considered bad, here the worst thing that could happen is that the
        # figure produced by the matplotlib backend upon closing is not
        # updated. Therefore, "except: pass" should be considered as an
        # acceptable workaround for this case.
        try:
            # redraw the figure, before closing it
            self.figure.canvas.draw()
        except:
            pass

    @property
    def figure(self):
        """
        The matplotlib figure that represents and/or adjusts the parameters of
        `self.modification`.
        """
        # Automatically initialize a figure
        if self._fig is None:
            self.init_fig(show=False)
        # Return a previously initialized figure
        return self._fig

    @figure.setter
    def figure(self, figure):
        self._fig = figure


class Modification(GraphMember, metaclass=ABCMeta):
    """
    Modification is an abstract class, that implements methods to modify the
    data of a `View` (`view_apply`) and adjust the parameters which control the
    behaviour of the modifications applied.

    Whenever one of the parameters needed to calculate the modification is
    changed, the view, this modification is applied to, is informed.
    `self.set_changed()` Has to be called upon any change of the modification
    that influences the behaviour of `self.modify()`. In essence, these are all
    parameters that are used to determine the modification. Therefore, this
    should be called by all setters of the parameters/attributes.

    Every subclass of Modification has to implement a constructor method
    `self.__init__(self, **kwargs)`, which calls the superclasses' constructor
    and sets the traces, the modification is applied to with the keyword
    parameter `traces_apply`. An example could be:
    super().__init__(traces_apply=['psdX', 'psdZ'], **kwargs)
    """
    # set a graphical modification, which will, per default, do nothing
    GRAPHICALMOD = GraphicalMod

    def __init__(self, traces_apply=None, view_apply=None, view_based=None,
                 automatic_switch=False, datapoints=-1, **kwargs):
        # Call the constructor of the superclass `GraphMember` and set the
        # maximum allowed number of parents (`view_based`) and childs
        # (`view_apply`) to one.
        super().__init__(max_children=1, max_parents=1, **kwargs)

        # A `Modification` has to be applied to a `View`!
        if view_apply is None:
            raise TypeError("Modification missing required positional argument"
                            " `view_apply`.")

        # Set the view, from where the parameters for the modification are
        # calculated from
        if view_based is not None:
            self.view_based = view_based

        # Set the view, whose data is going to be modified
        self.view_apply = view_apply

        # Set the traces, which are modified by this `Modification`
        self.traces_apply = traces_apply

        # Initialize InteractiveAttributes object, which will hold all the
        # parameters that the user should interact with.
        self.iattributes = InteractiveAttributes()

        # A checkbox to switch on/off the automatic determination of the
        # parameters that are used to calculate the modification in the method
        # `self.recalculate()`. The attribute `self.automatic` is checked in
        # the method `self.recalculate()`. If `automatic` is True, the
        # parameters are recalculated, otherwise the parameters are left
        # unchanged. Whenever `automatic` is changed (by the user or
        # automatically), `self.evaluate()` is called.
        if automatic_switch:
            self.add_iattribute('automatic', description='Automatic mode',
                                value=True, unset_automatic=False,
                                set_changed=False,
                                callback_functions=[self.evaluate])

        # A checkbox to de-/activate this `Modification`. This attribute gets
        # evaluated by `self.modify()`. If the `Modification` is active, it
        # modifies data, otherwise not, i.e. modify() returns modified or
        # unmodified original data, respectively.
        desc = "".join((self.__class__.__name__, " active"))
        self.add_iattribute('active', description=desc, value=True,
                            unset_automatic=False)

        # Datapoints is used to calculate and/or present modification. The
        # attribute `datapoints` is used to calculate a decimating factor and
        # speed up the calculations and/or plot commands.
        if datapoints > 0:
            desc = "Datapoints to calculate/visualize modification"
            self.add_iattribute('datapoints', description=desc,
                                value=datapoints, unset_automatic=False)

        # Add a Button to manually call the method `self.evaluate()`.
        self.add_iattribute('evaluate', description='Evaluate',
                            unset_automatic=False, set_changed=False,
                            callback_functions=[self.evaluate])

    def add_iattribute(self, key, description=None, value=None,
                       unset_automatic=True, set_changed=True,
                       callback_functions=None, **kwargs):
        """
        Add logic for automatic checkbox.
        Register widget with unset_automatic=True
        (-> Upon change of widget, unset automatic mode).
        Change default behaviour by setting kwarg: unset_automatic = False
        Add logic for triggering changed (calling self.set_changed).
        Register widget with set_changed=True.
        """
        if callback_functions is None:
            callback_functions = []

        if unset_automatic:
            callback_functions.append(self._unset_automatic)

        if set_changed:
            callback_functions.append(self.set_changed)

        self.iattributes.add(key, description=description, value=value,
                             callback_functions=callback_functions, **kwargs)

    def _unset_automatic(self, leave_automatic=False, **kwargs):
        """
        Add the logic for the automatic checkbox. If the value of an attribute
        is changed and the attribute was created with `unset_automatic=True`,
        deactivate the automatic mode (see `self.add_iattribute()`). To
        temporarily leave the automatic mode status untouched when changing the
        value of an attribute, i.e. not unset the automatic mode, set the value
        of the attribute with the keyword argument `leave_automatic=True`
        (see method `self.iattributes.set_value()`)
        """
        if not leave_automatic:
            self.iattributes.set_value('automatic', False, callback=False)

    def evaluate(self):
        """
        Implement the (re)calculation for the values necessary to calculate the
        modification in the subclass and call recalculate() of the superclass
        (this class).
        """
        if self.updated:
            # This method makes sure the modification is calculated with the
            # current values of the View this modification is based on. It is
            # called by self.modify().

            # When a View requests data, it calls modify(), which in turn calls
            # recalculate(). Recalculate(), if necessary, calls
            # get_data_modified() from the View it is based on, which again
            # triggers a call of modify() and a subsequent recalcaulte() of all
            # modifications associated with this View.
            # Modification need update, because view, this mod is based on,
            # was changed.
            # self._view_based.evaluate()is not needed, it is called via:
            # recalculate() -> get_data_based() -> _view_based.get_data() ->
            # get_modified_data() -> super().evaluate()
            return

        # Recalculate and print info of recalculated values if in automatic
        # mode
        if self.recalculate():
            self.print_info()

        # Update figure after recalculation has taken place
        self.graphicalmod.update()

    def recalculate(self):
        # Check if recalculation of parameters is necessary
        if self.updated:
            return False
        # Check the attribute self.automatic, whether the parameters needed for
        # the calculation of the modification should be determined
        # automatically or not. If values are set manually, no recalculation is
        # necessary, and `self` is therefore up to date.
        if not self.automatic:
            self.updated = True
            return True
        # Recalculate the parameters, inform the view this `Modification`
        # is applied to about the change, and set `self` to be updated.
        self._recalculate()
        self.set_changed(updated=True)
        return True

    def _recalculate(self):
        """
        This method should be overwritten by subclasses and perform the
        recalculation necessary to determine the parameters used by this
        Modification to modify the data in `self._modify()`.
        """
        pass

    def print_info(self):
            print("Values for Modification of class %s:"
                  % self.__class__.__name__)
            if not self.automatic:
                print("  Parameters set manually!")
            for key, widget in self.iattributes._widgets.items():
                if hasattr(widget, 'value'):
                    if isinstance(widget.value, float):
                        print("    %s: %.5f" % (widget.description, widget.value))
                    if isinstance(widget.value, collections.Iterable):
                        print("    %s: %s" % (widget.description, widget.value))
            self._print_info()

    def _print_info(self):
        """
        This method should be overwritten by subclasses, which want to print
        extra info additionally to the info of the calculated paremeters.
        """
        pass

    def modify(self, data, samples, traces_idx):
        """
        Modifies data and returns the modified array.

        Parameters
        ----------
        data : 2D numpy.ndarray of type float
            `data` holds the data to be modified
        samples : index array or slice
            `samples` is the index of the samples that was used to get the
            `data`
        traces : index array or slice
            `traces` is the index of the traces that was used to get the `data`
        """
        # Modification is active.
        if self.active:
            # Check if traces contained in data are modified by this
            # modification.
            data_traces = self.view_apply.idx_to_traces(traces_idx)
            mod_traces = self.traces_apply

            # Calculate the indices of traces contained in data and
            # modification. First, calculate indices of modification traces.
            mod_index = hp.overlap_index(mod_traces, data_traces)
            if len(mod_index) > 0:
                # At least one trace exists in both data and modification.
                # Therefore, the data needs to be modified...
                mod_index = hp.slicify(mod_index)

                # Calculate indices of traces of the data in such a way that
                # `data[:, data_index]` indexes the same traces as
                # `self.traces_apply[mod_index]`
                data_index = np.array([data_traces.index(trace)
                                       for trace
                                       in np.array(mod_traces)[mod_index]])
                data_index = hp.slicify(data_index)

                # Trigger a recalculation of the parameters for the
                # modification (if necessary) before modifying the data.
                self.evaluate()

                # Modify and return the modified data
                return self._modify(data=data,
                                    samples=samples,
                                    data_traces=data_traces,
                                    data_index=data_index,
                                    mod_index=mod_index)
        # Return unmodified data
        return data

    @abstractmethod
    def _modify(self, data, samples, data_traces, data_index, mod_index):
        """
        Is called by self.modify() whenever data is requested and needs to be
        modified.

        Parameters
        ----------
        data : 2D numpy.array()
            Contains the data, indexed by samples and data_traces
        samples : slice or 1D numpy.array()
            Is the index of the samples contained in data, which was
            given/asked by the user/process who called _get_data().
        data_traces : list of str
            Contains a list of traces (str) existent in data, which
            was given/asked by the user/process who called _get_data().
        data_index : slice or 1D numpy.array()
            data[:, data_index] gives the data, which is modified by
            this modification
        mod_index : slice or 1D numpy.array()
            np.array(self.traces_apply)[mod_index] gives the traces,
            which are existent in data and also modified by this modfication.

        Returns
        -------
        2D numpy.array()
            The modified data.
        """
        # modify data here, like so:
        # data[:,data_index] -= modification[:,mod_index]
        return data

    @property
    def updated(self):
        return self._updated

    @updated.setter
    def updated(self, value):
        """
        Gets set to True, after all `Views`, this `Modification` is based on,
        have been updated and after this `Modification` has been recalculated.
        This is automatically taken care of by `self.evaluate()` ->
        `self.recalculate()`.

        Gets called by a `View`, this `Modification` is based on, whenever the
        `View` (a `Modification` of the `View`) has been changed. It
        automatically informs its own `View`, that there was a change, by
        calling `self.set_changed()`.
        """
        self._updated = value

    def member_changed(self, ancestor=True, calledfromself=False,
                       index_shift=None, **kwargs):
        # If a change of an ancestor View or a MultiRegion was triggered by an
        # index_shift, the modification needs to recalculate itself, i.e.
        # the modification will alter its changeing behaviour. Because an
        # index_shift change is only transmitted to `level=1`, inform the
        # descendants of the change itself. A change of descendants is ignored.
        if index_shift is not None and not calledfromself and ancestor:
            self.set_changed(includeself=False)

        # Update update status
        super().member_changed(ancestor=ancestor,
                               calledfromself=calledfromself, **kwargs)

    def _get_data(self, based=True, samples=None, traces=None, window=False,
                  decimate=False, copy=True):
        if based:
            view = self.view_based
        else:
            view = self.view_apply

        if not isinstance(window, bool) and isinstance(window, int):
            window = window
        elif window:
            window = self.decimate
        else:
            window = 1

        if not isinstance(decimate, bool) and isinstance(decimate, int):
            decimate = decimate
        elif decimate:
            decimate = self.decimate
        else:
            decimate = 1

        if not based:
            old_active = self.iattributes.active
            self.iattributes.set_value('active', False, callback=False)

        data = view.get_data(traces=traces, samples=samples,
                             moving_filter='mean', window=window,
                             decimate=decimate, copy=copy)

        if not based:
            self.iattributes.set_value('active', old_active, callback=False)

        return data

    def _get_data_based(self, samples=None, traces=None, window=False,
                        decimate=False, copy=True):
        """
        decimate is False per default. If decimate is True, it only gets used,
        if samples are set to None (step information in samples precedes over
        decimate).
        """
        return self._get_data(based=True, samples=samples, traces=traces,
                              window=window, decimate=decimate, copy=copy)

    def _get_data_apply(self, samples=None, traces=None, window=False,
                        decimate=False, copy=True):
        """
        Get data of view apply with all modifications applied, except self.
        This is achieved by setting the self.__active flag to False.
        self.__active is intentionally set directly by accessing the attribute
        and not using the property/set_active() method, to prevent firing the
        self.set_changed() method within the set_active() method.
        decimate is False per default. If decimate is True, it only gets used,
        if samples are set to None (step information in samples precedes over
        decimate).
        """
        return self._get_data(based=False, samples=samples, traces=traces,
                              window=window, decimate=decimate, copy=copy)

    def calculate_bin_means(self, data=None, traces=None, bins=None,
                            datapoints_per_bin=None, sorttrace=0):
        """
        Calculates binned means based on the data to be fitted. The binned
        means are usually used by data fitting routines.

        Parameters
        ----------
        data : 2D numpy.ndarray of type float, optional
            Defaults to `self._get_data_based(traces=traces, decimate=True)`.
        traces : str or list of str, optional
            Defaults to `self.traces_apply`.
        bins : int, optional
            Number of bins that contain the datapoints to be averaged. If
            possible, it defaults to (`self.iattributes.datapoints` /
            `datapoints_per_bin`), otherwise bins defaults to
            (`self.view_based.datapoints` / `datapoints_per_bin`).
        datapoints_per_bin : int, optional
            Average number of datapoints to be averaged in one bin. Defaults to
            25.
        sorttrace : int, optional
            Trace (column) of `data` that acts as sorting index upon binning
            for the rest of the data. Defaults to the first trace of the data.

        Returns
        -------
        1D numpy.ndarray of type float
            The averaged bin values.
        float
            The size of one bin.
        """
        # Bin data and average bins to prevent arbitrary weighting of bins with
        # more datapoints
        if bins is None:
            bins = self._bins(datapoints_per_bin=datapoints_per_bin)

        # get the traces to retrieve data from
        if traces is None:
            traces = self.traces_apply

        # get the data to bin
        if data is None:
            data = self._get_data_based(traces=traces, decimate=True)

        # create the bins based on one trace of the data
        minimum = np.min(data[:, sorttrace])
        maximum = np.max(data[:, sorttrace])
        edges = np.linspace(minimum, maximum, bins + 1)

        # Get the indices of the bins to which each value in input array
        # belongs.
        bin_idx = np.digitize(data[:, sorttrace], edges)

        # Find which points are on the rightmost edge.
        on_edge = data[:, sorttrace] == edges[-1]
        # Shift these points one bin to the left.
        bin_idx[on_edge] -= 1

        # fill the bins with the means of the data contained in each bin
        bin_means = np.array([data[bin_idx == i].mean(axis=0)
                              for i in range(1, bins + 1)
                              if np.any(bin_idx == i)])

        bin_width = edges[1] - edges[0]

        return bin_means, bin_width

    def _bins(self, datapoints_per_bin=None):
        # On average 25 datapoints per bin
        datapoints_per_bin = datapoints_per_bin or 25
        if 'datapoints' in self.iattributes:
            bins = self.iattributes.datapoints / datapoints_per_bin
        else:
            bins = self.view_based.datapoints / datapoints_per_bin
        bins = max(1, int(np.round(bins)))
        return bins

    _NAME = {
        'position': ['positionX', 'positionY'],
        'psd': ['psdX', 'psdY'],
        'axis': ['X', 'Y']
    }

    def _excited(self, traces=None):
        traces = traces or ['positionX', 'positionY']
        data = self._get_data_based(traces=traces, copy=False)
        return sn.get_excited_signal(data)

    def interact(self):
        self.recalculate()
        self.iattributes.display()
        self.graphicalmod.display()

    @property
    def graphicalmod(self):
        # ZODB volatile
        if not hasattr(self, '_v_graphicalmod'):
            self._v_graphicalmod \
                = self.__class__.GRAPHICALMOD(modification=self)
        return self._v_graphicalmod

    @property
    def active(self):
        active = False
        if 'active' in self.iattributes:
            active = self.iattributes.active
        return active

    @property
    def automatic(self):
        # Does the modification automatically calculate its parameters
        automatic = True
        if 'automatic' in self.iattributes:
            automatic = self.iattributes.automatic
        return automatic

    @property
    def datapoints(self):
        if 'datapoints' in self.iattributes:
            return self.iattributes.datapoints
        else:
            return self.view_based.datapoints

    @property
    def decimate(self):
        if 'datapoints' in self.iattributes:
            return max(1, int(np.round(self.view_based.datapoints
                                       / self.datapoints)))
        else:
            return 1

    @property
    def view_based(self):
        return self.parent

    @property
    def view_apply(self):
        return self.child

    @view_based.setter
    def view_based(self, view):
        self.set_parent(view)

    @view_apply.setter
    def view_apply(self, view):
        self.set_child(view)

    def lia(self, trace):
        """
        Return the local index of trace in traces_apply
        """
        return self.traces_apply.index(trace)

    @property
    def traces_apply(self):
        # return a copy to protect local copy
        return self._traces_apply.copy()

    @traces_apply.setter
    def traces_apply(self, traces):
        if traces is None:
            traces_apply = []
        else:
            traces_apply = tc.normalize(traces)
        self._traces_apply = traces_apply



# -*- coding: utf-8 -*-
# """
# - Author: steve simmert
# - E-mail: steve.simmert@uni-tuebingen.de
# - Copyright: 2015
# """
from . import ureg

from . import name_constants as co

from .plotting import add_plot_to_figure
from .plotting import col_dict

from .psd import apply_aliasing
from .psd import apply_low_pass_filter
from .psd import lorentzian_psd
from .psd import hydro_psd

from .utilities import check_and_separate
from .utilities import flatten_list
from .utilities import gen_fit_pars
from .utilities import logspace_points_per_decade

from collections import OrderedDict

from configparser import ConfigParser

from functools import partial

from lmfit import minimize
from lmfit import Model
from lmfit import report_fit

import matplotlib.pyplot as plt

from numpy import nan

from os.path import join, isfile

from scipy import array
from scipy import complex_
from scipy import inf
from scipy import isnan
from scipy import logical_and
from scipy import pi
from scipy import sqrt

from scipy.constants import Boltzmann as k_B
from scipy.stats import norm

import sys

import time

import warnings


def gen_psd_fit_pars(D=1, D_min=0,
                     f_c=1000, f_c_min=0, f_c_max=1e7,
                     **kwargs):
    """
    Generate psd fit parameters.

    Every keyword will create a parameter, keywords ending with '_min' or
    '_max' will define the upper and lower bound for the parameter.

    Example
    -------
    gen_fit_pars(D=1e-6, D_min=0, f_c=1000, f_c_max=1e6) will create
    OrderedDict([('D', <Parameter 'D', 1e-06, bounds=[0:None]>),
    ('f_c', <Parameter 'f_c', 1000, bounds=[None:1000000.0]>)]).

    Returns
    -------
    params : Parameters
        lmfit Parameters object
    """
    params = gen_fit_pars(D=D, D_min=D_min,
                          f_c=f_c, f_c_min=f_c_min, f_c_max=f_c_max,
                          **kwargs)
    return params


def gen_model_fun(model='lorentzian',
                  lp_filter=False,
                  lp_fixed=False,
                  aliasing=False,
                  f_sample=None,
                  N_alias=9,
                  model_fun=None,
                  param_names=None,
                  name='other',
                  invert_function=False,
                  **kwargs):
    """
    Generate the model function.

    invert function has no effect on userdefined model_functions.
    They must be inverted already!
    """
    ################################
    # get the basic model function #
    ################################
    model = model.lower()
    if model == 'lorentzian':
        model_fun = lorentzian_psd
        name = 'Lorentzian PSD'
    elif model == 'hydro':
        model_fun = hydro_psd
        name = 'Hydr. PSD'
    elif model == 'other':
        if model_fun is None:
            raise Exception('Function missing in keyword argument model_fun')
    else:
        raise Exception('Model unknown model={}'.format(model))

    ###################################
    # define the fitting parameters #
    ###################################
    if param_names is None:
        param_names = ['D', 'f_c']
        if lp_filter:
            name += ', low-pass filtered'
            param_names.append('f3dB')
            param_names.append('alpha')

    # get rid of init guess kwargs, which are not important here
    for pname in param_names:
        try:
            kwargs.pop(pname)
        except:
            pass
    #################################
    # generate the fitting function #
    #################################
    # separate the function keyword arguments from all keyword arguments
    model_kwargs = check_and_separate(model_fun, kwargs, exclude=True)

    # basic model function (partial), e.g. hydro_psd(freq, D, f_c)
    pfun = partial(model_fun, **model_kwargs)
    pfun.__name__ = model_fun.__name__
    if not model_fun.__doc__:
        model_fun.__doc__ = ''
    pfun.__doc__ = ('partial of {0:s}\nOriginal doc\n{1:s}'
                    ''.format(pfun.__name__, model_fun.__doc__))

    # if user defined function is provided don't do aliasong etc.
    # this must have been don by the user beforehand
    if model != 'other':
        # define function w/ resp. to lp_filter and aliasing and INVERT
        if aliasing:
            name += ' and aliased'

            if f_sample is None:
                raise Exception('Sampling frequency is missing.')

            if lp_filter:
                def aliased_lp_fun(freq, D, f_c, f3dB, alpha):
                    f = apply_low_pass_filter(pfun, f3dB, alpha)
                    f_ = apply_aliasing(f, f_sample, N_alias)(freq, D, f_c)
                    if invert_function:
                        return 1 / f_
                    else:
                        return f_

                model_fun_ = aliased_lp_fun
            else:
                def aliased_fun(freq, D, f_c):
                    f = apply_aliasing(pfun, f_sample, N_alias)(freq, D, f_c)
                    if invert_function:
                        return 1 / f
                    else:
                        return f

                model_fun_ = aliased_fun
        else:
            if lp_filter:
                def lp_fun(freq, D, f_c, f3dB, alpha):
                    f = apply_low_pass_filter(pfun, f3dB, alpha)(freq, D, f_c)
                    if invert_function:
                        return 1 / f
                    else:
                        return f

                model_fun_ = lp_fun
            else:
                def fun(freq, D,  f_c):
                    if invert_function:
                        return 1 / pfun(freq, D, f_c)
                    else:
                        return pfun(freq, D, f_c)

                model_fun_ = fun
    else:
        model_fun_ = pfun

    return (model_fun_, param_names, name)


def fit_psd(freq,
            psd_data,
            psd_err=None,
            N_avg=1,
            model='lorentzian',
            lp_filter=False,
            lp_fixed=False,
            aliasing=False,
            f_sample=None,
            N_alias=9,
            model_fun=None,
            name='other',
            verbose=False,
            **kwargs):
    """
    Fits the given power spectral density to the selected model.

    kwargs holds both model_kwargs and inital guess and min max kwargs.
    """
    #################################
    # generate the fitting function #
    #################################
    # kwargs still holds both, model_kwargs and inital guesses, min max etc.
    model_fun_, param_names, name = gen_model_fun(model=model,
                                                  invert_function=True,
                                                  lp_filter=lp_filter,
                                                  lp_fixed=lp_fixed,
                                                  aliasing=aliasing,
                                                  f_sample=f_sample,
                                                  N_alias=N_alias,
                                                  model_fun=model_fun,
                                                  name=name,
                                                  **kwargs)
    # kwargs should only have the initial guesses and min, max etc.

    ###################################
    # generate the fitting parameters #
    ###################################
    # get initial guesses and mins and maxs from kwargs or create them
    par_kwargs = {}
    suffix = ['', '_min', '_max']
    for pname in param_names:
        for suff in suffix:
            try:
                par_kwargs[pname + suff] = kwargs.pop(pname + suff)
            except:
                pass
    if lp_filter:
        if 'f3dB' not in par_kwargs:
            par_kwargs['f3dB'] = 10e3
            if lp_fixed:
                warnings.warn('cut-off frequency not provided! '
                              'Using 10 kHz as fallback.')
        if 'f3dB_min' not in par_kwargs:
            par_kwargs['f3dB_min'] = 0

        if 'alpha' not in par_kwargs:
            par_kwargs['alpha'] = 0.5
            if lp_fixed:
                warnings.warn('alpha not provided! Using 0.5 as fallback.')
        if 'alpha_min' not in par_kwargs:
            par_kwargs['alpha_min'] = 0
        if 'alpha_max' not in par_kwargs:
            par_kwargs['alpha_max'] = 1

    # generate the Parameters object
    pars = gen_psd_fit_pars(**par_kwargs)

    if lp_fixed:
        pars['f3dB'].vary = False
        pars['alpha'].vary = False

    ############################
    # generate the lmfit.Model #
    ############################
    model_obj = Model(model_fun_,
                      independent_vars=['freq'],
                      param_names=param_names,
                      name=name)

    #################################
    # calculate the weights and fit #
    #################################
    if psd_err is None:
        # err = psd_data / sqrt(N_avg)
        # err_i = 1 / (psd_data * sqrt(N_avg))
        weights = psd_data * sqrt(N_avg)
    else:
        # err = psd_data / sqrt(N_avg)
        # err_i = 1 / (psd_data * sqrt(N_avg)) = 1 / (psd_err * N_avg)
        # err_i = 1 / (err * N_avg)
        weights = psd_err * N_avg

    modelfit = model_obj.fit(1/psd_data, freq=freq,
                             params=pars, weights=weights,
                             nan_policy='propagate')
    if verbose:
        report_fit(modelfit)

    return modelfit


def analytical_lsq_lorentzian(freq, psd, verbose=False):
    """
    Analytically solve the least-squares problem of a Lorentzian-shaped power
    spectrum.

    Calculates the least-squares fit of a Lorentzian power spectrum
    analytically for the given frequencies and power spectral density values
    (see Equs. 13 and 14 in Berg-Sørensen & Flyvbjerg (2004))

    Arguments
    ---------
    freq : array(float)
        Frequency vector.
    psd : array(float)
        psd values.
    verbose : bool
        Will give additional information, if True.

    Returns
    -------
    f_c : float
        The corner frequency
    D : float
        The diffusion coefficient
    chi2 : float
        The reduced chi-squared value of the fit.

    Note
    ----
    Bias
        A biased diffusion constant is NOT accounted for as it is stated in
        Equ. (41) in Nørrelykke & Flyvbjerg (2010).

    Eaution
        A factor two is omitted in the formula for D because we are dealing with
        one-sided power spectra.

    References
    ----------
    Berg-Sørensen, K., & Flyvbjerg, H. (2004)
        Equ. (13, 14) in
        Power spectrum analysis for optical tweezers. Review of Scientific
        Instruments, 75(3), 594–612.
        http://doi.org/10.1063/1.1645654

    Nørrelykke & Flyvbjerg (2010)
        Power spectrum analysis with least-squares fitting:
        Amplitude bias and its elimination, with application to optical
        tweezers and atomic force microscope cantilevers.
        Review of Scientific Instruments, 81(7), 075103.
        http://doi.org/10.1063/1.3455217
    """
    def ls_sum(freq, psd, p, q):
        return sum(freq**(2 * p) * psd**q)

    S_00 = ls_sum(freq, psd, 0, 0)
    S_01 = ls_sum(freq, psd, 0, 1)
    S_02 = ls_sum(freq, psd, 0, 2)
    # S_10 = ls_sum(freq, psd, 1, 0)
    S_11 = ls_sum(freq, psd, 1, 1)
    S_12 = ls_sum(freq, psd, 1, 2)
    # S_21 = ls_sum(freq, psd, 2, 1)
    S_22 = ls_sum(freq, psd, 2, 2)

    a = (S_01 * S_22 - S_11 * S_12)
    b = (S_11 * S_02 - S_01 * S_12)

    f_c = sqrt(a / b).real
    D = pi * pi * (S_02 * S_22 - S_12 * S_12) / (S_11 * S_02 - S_01 * S_12)
    chi2 = S_00 - ((S_01 * S_01 * S_22 +
                    S_11 * S_11 * S_02 - 2 * S_01 * S_11 * S_12) /
                   (S_02 * S_22 - S_12 * S_12))

    if isnan(f_c):
        if verbose:
            warnings.warn('Corner frequency calculation resulted in NAN.'
                          'f_c = (({0:1.17e} / {1:1.17e})**0.5).real'
                          ''.format(a, b))
            print('The following sums were calculated:'
                  'S_00={}\t S_01={}\t S_02={}\n'
                  'S_11={}\t S_12={}\t S_22={}'
                  ''.format(S_00, S_01, S_02, S_11, S_12, S_22))
            print('with fmin = {}, fmax={}'.format(min(freq), max(freq)))
            print('a=(S_01 * S_22 - S_11 * S_12)')
            print('b=(S_11 * S_02 - S_01 * S_12)')

    return (f_c, D, chi2)


def calc_anal_lsq_opt(freq, psd, verbose=False, plot=False, ppd=20):
    """
    Try to find the optimal range for the analytical least-squares algorithm,
    by varying the range of the frequency vector.

    The function varies the range of data points to find an optimal least-
    squares fit of the data. this is done as follows:
    First, the highest frequency boundary is varied, while the lower is kept
    fixed. For each new set of lower and upper bound the
    *analytical_lsq_lorentzian*-method is called. The frequency, at which the
    chi2 value is minimal, is takes as upper bound.
    Second, the upper frequency bound is varied and the frequency, at which the
    corner frequency is highest is taken as lower bound.

    Arguments
    ---------
    freq : array(float)
        Frequency vector.
    psd : array(float)
        psd values.
    verbose : bool
        Will give additional information, if True.
    plot : bool
        Whether, to plot the trials (debugging)
    ppd : int
        Number of point per decade. The more the higer the number of trials
        (default is 20).

    Returns
    -------
    f_c : float
        The corner frequency of the *best* fit.
    D : float
        The diffusion coefficient of the *best* fit.
    chi2 : float
        The reduced chi-squared value of the *best* fit.
    (fmin, fmax) : tuple
        Tuple defining the boundaries of the *best* fit.

    See Also
    --------
    analytical_lsq_lorentzian
    """
    if verbose:
        print('Trying to find optimal boundaries for the analytical least '
              'squares of a Lorentzian.')
        print('Number of points per decade: {}'.format(ppd))

    fmin_init = min(freq)

    # decimate frequency points
    #f = logspace(log10(min(freq)),
    #             log10(max(freq)),
    #             base=10, num=n_trials)
    f = logspace_points_per_decade(min(freq), max(freq), ppd=ppd)
    # define a set of test values for fmin and fmax
    fmin_ = f[:int(len(f)/2)]
    fmax_ = f[int(len(f)/2):]

    if verbose:
        print('initial fmin: {0:6.2f}'.format(fmin_init))
        print('fmin - fmax border: {0:6.2f}'.format(f[round(len(f) / 2)]))

    # vary the upper freq bound and get corner frequency and chi² value
    # from the analytical least squares calculation
    fc2 = []
    ch2 = []
    for fmax in fmax_:
        excl = ((freq < fmin_init) + (freq > fmax))
        keep = ~excl

        freq_part = freq[keep]
        res = analytical_lsq_lorentzian(freq_part, psd[keep], verbose=verbose)
        if not(any(isnan(a) for a in res)):
            fc2.append(res[0])
            ch2.append(res[2] / len(freq_part))

    # set the upper bound where chi² is minimal
    if len(ch2) > 0:
        fmax = array(fmax_)[ch2 == min(ch2)][0]
    else:
        fmax = max(freq)
    # plot the corner frequencies and chi² values vs the upper
    # frequency bound
    if plot:
        fig = plt.figure()
        ax1 = add_plot_to_figure(fig, [fmax, fmax], [max(ch2), min(ch2)],
                                 subplot=(1, 3, 1))
        ax1 = add_plot_to_figure(fig, fmax_, fc2,
                                 title=(r'$\mathsf{f_c and \chi^2 vs. '
                                        'f_{max}}$'),
                                 ylabel=r'$\mathsf{f_c (Hz)}$',
                                 fmt='or',
                                 axis=ax1,
                                 label=r'$\mathsf{f_c}$',
                                 showLegend=True,
                                 legend_kwargs={'loc': 2})
        ax1a = ax1.twinx()
        add_plot_to_figure(fig, fmax_, ch2,
                           xlabel=r'$\mathsf{f_{max}}$',
                           ylabel=r'$\mathsf{\chi^2}$',
                           fmt='og', axis=ax1a,
                           label=r'$\mathsf{\chi^2}$',
                           showLegend=True)
        #while not plt.waitforbuttonpress():
        #    pass
        #ax.figure.clf()

    if verbose:
        print('optimal fmax: {0:1.3e}'.format(fmax))

    # now vary the lower frequency bound with fixed ('optimal')
    # upper  bound. Calculate again the analytical LSq
    fc1 = []
    ch1 = []
    for fmin in fmin_:
        excl = ((freq < fmin) + (freq > fmax))
        keep = ~excl

        freq_part = freq[keep]
        res = analytical_lsq_lorentzian(freq_part, psd[keep])

        if not(any(isnan(a) for a in res)):
            fc1.append(res[0])
            ch1.append(res[2] / len(freq_part))
        else:
            fc1.append(-1.0)
            ch1.append(-1.0)

    # take the lower bound where the corner freq is highest
    if len(fc1) > 0:
        fmin = array(fmin_)[fc1 == max(fc1)][0]
    else:
        fmin = fmin_init
    # plot the corner frequencies and the chi² value vs lower freq bound
    if plot:
        fig = plt.figure()
        ax2 = add_plot_to_figure(fig, [fmin, fmin], [max(fc1), min(fc1)],
                                 subplot=(1, 3, 2))
        add_plot_to_figure(fig, fmin_, fc1,
                           ylabel=r'$\mathsf{f_c (Hz)}$',
                           fmt='or',
                           axis=ax2,
                           label=r'$\mathsf{f_c}$',
                           legend_kwargs={'loc': 2},
                           showLegend=True)
        ax2a = ax2.twinx()
        add_plot_to_figure(fig, fmin_, ch1,
                           title=r'$\mathsf{f_c and \chi^2 vs. f_{min}}$',
                           xlabel=r'$\mathsf{f_{min}}$',
                           ylabel=r'$\mathsf{\chi^2}$',
                           fmt='og', axis=ax2a,
                           label=r'$\mathsf{\chi^2}$',
                           showLegend=True)
        #while not plt.waitforbuttonpress():
        #    pass
        #ax.figure.clf()

    if verbose:
        print('optimal fmin: {0:6.2f}'.format(fmin))

    excl = ((freq < fmin) + (freq > fmax))
    keep = ~excl
    freq_part = freq[keep]
    f_c, D, chi2 = analytical_lsq_lorentzian(freq_part, psd[keep])

    if verbose:
        print('(f_c, D, chi2) = ({0:1.2f}, {1:1.3e}, {2:1.2f})'
              ''.format(f_c, D, chi2))

    if plot:
        fig = plt.figure()
        ax3 = add_plot_to_figure(fig, freq, psd,
                                 label='PSD',
                                 fmt='+',
                                 alpha=0.3,
                                 subplot=(1, 3, 3)
                                 )
        analpsd = lorentzian_psd(freq_part, D, f_c)
        add_plot_to_figure(fig, freq_part, analpsd,
                           axis=ax3,
                           title='PSD and analytical Lorentzian fit',
                           xlabel=r'Frequency',
                           ylabel=r'$\mathsf{PSD \; (V^2/Hz)}$',
                           label='Lorentzian fit',
                           fmt='--r',
                           showLegend=True,
                           logplot=True
                           )
    return (f_c, D, chi2, (fmin, fmax))


def collective_psd_fit_fun(params,
                           names,
                           freq_dict,
                           model_fun,
                           param_names,
                           psd_dict=None,
                           err_dict=None,
                           verbose=False
                           ):
    """
    residual function for teh collective psd fit.
    """
    psd_ = []
    err_ = []
    model_ = []

    for name in names:
        kws = {pname: params[pname + '_' + name].value
               for pname in param_names if pname not in ['f3dB', 'alpha']}
        kws['f3dB'] = params['f3dB'].value
        kws['alpha'] = params['alpha'].value

        model_.append(list(model_fun(freq_dict[name], **kws)))

        if psd_dict is not None:
            psd_.append(list(psd_dict[name]))
        if err_dict is not None:
            err_.append(list(err_dict[name]))

    model_flat_array = array(list(flatten_list(model_)))

    if psd_dict is not None:
        psd_flat_array = array(list(flatten_list(psd_)))

    if err_dict is not None:
        err_flat_array = array(list(flatten_list(err_)))

    if psd_dict is None:
        return model_flat_array
    elif err_dict is None:
        return (psd_flat_array - model_flat_array)
    else:
        return ((psd_flat_array - model_flat_array) / err_flat_array)


def make_collective_psd_fit(freq_dict,
                            psd_dict,
                            err_dict,
                            model='lorentzian',
                            aliasing=False,
                            f_sample=None,
                            N_alias=9,
                            verbose=False,
                            **kwargs):
    """
    Fit the provided PSDs collectively, providing the ability to fit low-pass
    filter parameters (alpha and f3dB ) same for all PSDs.
    Returns <lmfit.Minimizer>

    Arguments
    ---------
    freq_dict : dict
        Dictionary of frequency vectors for psds with name "name".
    psd_dict : dict
        Dictionary with names as keys and psds as values.
        If None, the function returns the function values instead of the
        residuals.
    err_dict : dict
        Dictionary with names as keys and psds as values.
        If None, the function returns unweighted residuals, i.e. physical
        errors.
    hydro_kwargs : dict
        Dictionary with keyword-arguments passed to hydro_psd(). I.e.
        {'R': radius, 'height': height, 'temp': Temperature,
        'rho': mass_density}
    params : lmfit.Parameters
        Object of the lmfit.Parameters class, holding Paramerter-objects with
        names 'D_name' and 'f_c_name' and the parameters for the low-pass
        filter 'f3dB' and 'alpha'. If None, they are created.
    init_vals : dict
        Dictionary with 'names' as keys and tuples (D, f_c) of initial
        guess-values for the respective psd fit.
        If None D = 1 and f_c = 1000 Hz is used.

    Return
    ------
    lmfit.Minimizer
    """
    names = psd_dict.keys()

    model_fun, param_names, name = gen_model_fun(model=model,
                                                 lp_filter=True,
                                                 lp_fixed=False,
                                                 aliasing=aliasing,
                                                 f_sample=f_sample,
                                                 N_alias=N_alias,
                                                 **kwargs)

    # get initial guesses and mins and maxs from kwargs or create them
    par_kwargs = {}
    suffix = ['', '_min', '_max']
    for pname in param_names:
        if pname in ['f3dB', 'alpha']:
            for suff in suffix:
                try:
                    par_kwargs[pname + suff] = kwargs.pop(pname + name)
                except:
                    pass
        else:
            for name in names:
                for suff in suffix:
                    try:
                        key = kwargs.pop(pname + '_' + name)
                        par_kwargs[pname + '_' + name + suff] = key
                    except:
                        if suff == '':
                            par_kwargs[pname + '_' + name] = 1

    if 'f3dB' not in par_kwargs:
        par_kwargs['f3dB'] = 10e3
    if 'f3dB_min' not in par_kwargs:
        par_kwargs['f3dB_min'] = 0

    if 'alpha' not in par_kwargs:
        par_kwargs['alpha'] = 0.5
    if 'alpha_min' not in par_kwargs:
        par_kwargs['alpha_min'] = 0
    if 'alpha_max' not in par_kwargs:
        par_kwargs['alpha_max'] = 1

    params = gen_fit_pars(**par_kwargs)

    minimizer = minimize(collective_psd_fit_fun, params,
                         args=(names, freq_dict, model_fun, param_names),
                         kws={'psd_dict': psd_dict,
                              'err_dict': err_dict,
                              'verbose': verbose})

    minimizer.fun = model_fun
    minimizer.param_names = param_names
    minimizer.fun_name = name

    if verbose:
        report_fit(minimizer)

    return minimizer


def is_outlier(red_chi2, N_free, cl=0.95):
    """
    Evaluates the reduced chi² values of the PSD fit and return true if the
    value is considered to be an outlier or false if not.

    Arguments
    ---------
    red_chi2 : array(float) or float
        Numpy array of reduced chi² values. The reduced chi² value is
        calculated as: red_chi² = chi² / N_free
    N_free : array(int) or int
        Number of degrees of freedom.
    conf_level : float
        Confidence level that specifies the range of reduced chi^2 values that
        you are confident about that they come from a normal distribution.

    Returns
    -------
    array of bool

    Note
    ----
    Short
        Say the confidence level is set to 0.95. That means you define every
        datapoint above the 0.975 and below the 0.025 quantile to be an
        outlier. This also means, you expect that, on average, every 20th
        measurement to be an outlier, even if it origins from the expected
        distribution.

    Longer
        In general, the probability that a value lies outside a given
        confidence interval is given by p_out = 1 - conf_level = 2 (1 - p),
        hence p = (1+conf_level)/2. Which, for conf_level = 0.95 means
        p = 0.975.

        The reduced chi2-value (chi2-value divided by the number of degrees of
        freedom) is distributed around a mean = 1; its standard deviation is
        SD = sqrt(2).

        For a large number of degrees of freedom N, the chi2-distribution is a
        normal distribution. The standard error of the mean of normally
        distributed data scales with the square-root of the number of
        measurements, i.e. SEM ~ 1/sqrt(N). Thus, we can ask for the quantile
        for p=0.975 of a normal distribution with mean = 1 and SD = sqrt(2/N)
    """
    # get the quantiles and calculate the upper and lower
    # limit of the reduced chi² values for p = (1+conf_level)/2
    upper_limit = norm.ppf((cl + 1) / 2,
                           loc=1,
                           scale=sqrt(2/N_free)
                           )
    lower_limit = 1 - (upper_limit - 1)
    outlier = (red_chi2 < lower_limit) + (red_chi2 > upper_limit)

    return outlier


class FitResult(object):
    def __init__(self,
                 name=None,
                 D=None,
                 D_err=None,
                 D_unit='V**2/s',
                 f_c=None,
                 f_c_err=None,
                 freq=None,
                 freq_unit='Hz',
                 eval=None,
                 psd_unit='V**2/Hz',
                 residuals=None,
                 chi2=None,
                 redchi2=None,
                 conf_level=None,
                 nfree=None,
                 params=None,
                 minimizer=None,
                 model=''
                 ):
        self.name = name
        self.D = D
        self.D_err = D_err
        self.D_unit = D_unit
        self.f_c = f_c
        self.f_c_err = f_c_err
        self.freq = freq
        self.freq_unit = freq_unit
        self.eval = eval
        self.psd_unit = psd_unit
        self.residuals = residuals
        self.chi2 = chi2
        self.redchi2 = redchi2
        self.conf_level = conf_level
        self.nfree = nfree
        self.params = params
        self.minimizer = minimizer
        self.model = model

    def get_freq_bounds(self):
        return (min(self.freq), max(self.freq))

    def get_D(self, unit='V**2/s'):
        if unit == 'V**2/s':
            conv = 1.0
        else:
            conv = ureg(self.D_unit).to(unit).magnitude
        return self.D * conv

    def get_f_c(self, unit='Hz'):
        if unit == 'Hz':
            conv = 1.0
        else:
            conv = ureg(self.freq_unit).to(unit).magnitude
        return self.f_c * conv

    def is_outlier(self, conf_level=None):
        if not conf_level:
            conf_level = self.conf_level

        return is_outlier(self.redchi2, self.nfree, cl=conf_level)

    def plot_fit(self, axis=None, **kwargs):
        """ plots the fit """
        ax = axis
        if ax is not None:
            fig = ax.figure
        else:
            fig = None

        l = ['plot_all', 'plot_masked', 'plot_errors']
        for key in l:
            try:
                kwargs.pop(key)
            except:
                pass

        if 'color' not in kwargs and self.name in col_dict.keys():
            kwargs['color'] = col_dict[self.name]
        if 'fmt' not in kwargs.keys():
            kwargs['fmt'] = '-'
        if 'linewidth' not in kwargs.keys():
            kwargs['linewidth'] = 1.5
        if 'alpha' not in kwargs.keys():
            kwargs['alpha'] = 0.7
        if 'showLegend' not in kwargs.keys():
            kwargs['showLegend'] = True
        if 'legend_kwargs' not in kwargs:
            lg_kws = {'loc': 3}
            kwargs['legend_kwargs'] = lg_kws
        if 'label' not in kwargs.keys():
            kwargs['label'] = '{} fit to {}'.format(self.model, self.name)
        if 'fontsize' not in kwargs.keys():
            kwargs['fontsize'] = 16

        ax = add_plot_to_figure(fig, self.freq, self.eval,
                                axis=ax, logplot=True, **kwargs)
        ax.grid(which='major')
        plt.setp(ax,
                 xlabel='Frequency (Hz)',
                 ylabel=r'PSD ($\mathsf{V^2/Hz}$)'
                 )
        return ax.figure


class Result(object):
    def __init__(self,
                 dissens,
                 dissens_err,
                 dissens_unit,
                 trap_stiffness,
                 trap_stiffness_err,
                 trap_stiffness_unit,
                 drag,
                 drag_err,
                 drag_unit,
                 excited=None,
                 ):
        self.dissens = dissens
        self.dissens_err = dissens_err
        self.dissens_unit = dissens_unit
        self.trap_stiffness = trap_stiffness
        self.trap_stiffness_err = trap_stiffness_err
        self.trap_stiffness_unit = trap_stiffness_unit
        self.drag = drag
        self.drag_err = drag_err
        self.drag_unit = drag_unit
        self.excited = excited

    def get_dissens(self, unit='nm/mV'):
        if unit != self.dissens_unit:
            conv = ureg(self.dissens_unit).to(unit).magnitude
        else:
            conv = 1.0
        return self.dissens * conv

    def get_dissens_err(self, unit='nm/mV'):
        if unit != self.dissens_unit:
            conv = ureg(self.dissens_unit).to(unit).magnitude
        else:
            conv = 1.0
        return self.dissens_err * conv

    def get_trap_stiffness(self, unit='pN/nm'):
        if unit != self.trap_stiffness_unit:
            conv = ureg(self.trap_stiffness_unit).to(unit).magnitude
        else:
            conv = 1.0
        return self.trap_stiffness * conv

    def get_trap_stiffness_err(self, unit='pN/nm'):
        if unit != self.trap_stiffness_unit:
            conv = ureg(self.trap_stiffness_unit).to(unit).magnitude
        else:
            conv = 1.0
        return self.trap_stiffness_err * conv

    def get_drag(self, unit='nN*s/m'):
        if unit != self.drag_unit:
            conv = ureg(self.drag_unit).to(unit).magnitude
        else:
            conv = 1.0
        return self.drag * conv

    def get_drag_err(self, unit='nN*s/m'):
        if unit != self.drag_unit:
            conv = ureg(self.drag_unit).to(unit).magnitude
        else:
            conv = 1.0
        return self.drag_err * conv

    def get_force_factor(self, unit='pN/mV'):
        """TODO"""
        beta = self.get_dissens(unit='nm/mV')
        kappa = self.get_trap_stiffness(unit='pN/nm')
        if unit != 'pN/mV':
            conv = ureg('pN/mV').to(unit).magnitude
        else:
            conv = 1.0
        alpha = beta * kappa * conv
        return alpha

    def get_force_factor_err(self, unit='pN/mV'):
        """ TODO """
        beta = self.get_dissens(unit='nm/mV')
        kappa = self.get_trap_stiffness(unit='pN/nm')
        beta_err = self.get_dissens_err(unit='nm/mV')
        kappa_err = self.get_trap_stiffness_err(unit='pN/nm')
        alpha = self.get_force_factor(unit=unit)

        alpha_err = (beta_err/beta + kappa_err/kappa) * alpha
        return alpha_err


class PSDFit(object):
    def __init__(self,
                 psdm,
                 bounds=None,
                 conf_level=0.95,
                 verbose=False,
                 **kwargs):
        """
        PSDFit object manages the several fitting procedures of power spectral
        densities of one measurement.

        When initialized, the *analytical_lorentzian_fit* method is called,
        which tries to find optimal bounds and initial values for the fits.

        Arguments
        ---------
        psdm : PSDMeasurement
            Object holding the power spectral densities of x, y and z.
        bounds : (fmin, fmax) or dict {name: (fmin, fmax), ...}
            defines the lower and upper frequency bound. All data outside the
            interval are masked. A dict specifies the bound for particular
            axes. The tuple will put the same boundary to each spectrum.
        kwargs
            Keyword arguments handed over to **analytical_lorentzian_fit()**.
        """
        self.psdm = psdm

        self.active_calibration = psdm.active_calibration
        self.names = psdm.names

        self.anal_fits = {}

        self.fit_kwargs = {}
        self.fits = {}
        
        self.conf_level = conf_level

        if bounds is not None:
            self.set_bounds(bounds)
        self.analytical_lorentzian_fit(verbose=verbose, **kwargs)

        self.pc_results = {}
        self.ac_results = {}

        self.exclude_freq_outside = self.set_bounds

    def __getattr__(self, name):
        """
        Shortcut for PSDM.attribute
        """
        if hasattr(self.psdm, name):
            return getattr(self.psdm, name)
        else:
            raise AttributeError(name)

    @property
    def freq_x(self):
        return self.get_freq('x')

    @property
    def freq_y(self):
        return self.get_freq('y')

    @property
    def freq_z(self):
        return self.get_freq('z')

    def get_freq(self, name, **kwargs):
        """
        Return the frequency vector of the psd with 'name'.

        Arguments
        ---------
        name : str
            name of the axis.
        kwargs can be:
        unit : str
        get_all : bool
            If False uses the internal mask.
        get_masked : bool
            If True returns the masked instead of the unmasked elements.
        """
        return self.psdm.get_freq(name=name, **kwargs)

    @property
    def psd_x(self):
        return self.get_psd('x')

    @property
    def psd_y(self):
        return self.get_psd('y')

    @property
    def psd_z(self):
        return self.get_psd('z')

    def get_psd(self, name, **kwargs):
        """
        Return the psd vector of the psd with 'name'.

        Arguments
        ---------
        name : str
            name of the axis.
        kwargs can be:
        unit : str
        get_all : bool
            If False uses the internal mask.
        get_masked : bool
            If True returns the masked instead of the unmasked elements.
        """
        return self.psdm.get_psd(name=name, **kwargs)

    @property
    def err_x(self):
        return self.get_psd_err('x')

    @property
    def err_y(self):
        return self.get_psd_err('y')

    @property
    def err_z(self):
        return self.get_psd_err('z')

    def get_psd_err(self, name, **kwargs):
        """
        Return the error vector of the psd with 'name'.

        Arguments
        ---------
        name : str
            name of the axis.
        kwargs can be:
        unit : str
        get_all : bool
            If False uses the internal mask.
        get_masked : bool
            If True returns the masked instead of the unmasked elements.
        """
        return self.psdm.get_psd_err(name=name, **kwargs)

    def set_bounds(self, bounds, names=None, reset_mask=False):
        """
        Set the frequency boundaries (fmin, fmax) at which to consider the psd
        values.

        Arguments
        ---------
        bounds : tuple (fmin, fmax) or dict {'name': (fmin, fmax), ...}
            Defines the lower and upper frequency bound. All data outside the
            interval is masked.
        names : list of strings
            Names to apply the tuple bounds to.
        reset : bool
            Resets the mask of the psd.
        """
        if hasattr(bounds, 'keys'):
            for name in bounds:
                fmin, fmax = bounds[name]
                self.psdm.exclude_freq_outside(fmin, fmax, names=name,
                                               reset_mask=reset_mask)
        else:  # either use bounds for all names or for 'names'
            if names and not isinstance(names, list):
                names = [names]

            for name in names if names else self.names:
                fmin, fmax = bounds
                self.psdm.exclude_freq_outside(fmin, fmax, names=name,
                                               reset_mask=reset_mask)

    def exclude_freq(self, f_exclude, names=None):
        """
        Exclude frequencies.

        Arguments
        ---------
        f_exclude : float or list of floats or dictionary that specifies the
            frequencies for each name. 'names' then has no effect.
        names : list of str
            Names to apply the exclusion of the list or float 'f_exclude'.
        """
        if hasattr(f_exclude, 'keys'):
            for name in f_exclude:
                self.psdm.exclude_freq(f_exclude[name], names=name)
        else:
            if names and not isinstance(names, list):
                names = [names]

            self.psdm.exclude_freq(f_exclude, names=names)

    def analytical_lorentzian_fit(self, names=None, verbose=False, **kwargs):
        """
        Calculates the analytical solution to the least squares problem of a
        simple Lorentzian.

        This function calls *calc_anal_lsq_opt()* to find a frequency interval
        where the data fits a lorentzian well.

        Arguments
        ---------
        names : [str, ...] or 'all'
            Defines for which PSD(s) the analytical lorentzian is calculated.
            Default is 'all'.
        verbose : bool
            Give more info.

        Keyword Arguments
        -----------------
        Keyword arguments passed over to *calc_anal_lsq_opt()*.

        References
        ----------
        K. Berg-Sørensen and H. Flyvbjerg
            Power spectrum analysis for optical tweezers
            Rev. Sci. Instr., vol. 75, 2004
        S.F. Nørrelykke and H. Flyvbjerg
            Power spectrum analysis with least-squares fitting: Amplitude bias
            and its elimination, with application to optical tweezers and
            atomic force microscope cantilevers
            Rev. Sci. Instr.,vol. 81, 2010

        See Also
        --------
        calc_anal_lsq_opt
        """
        if names and not isinstance(names, list):
            names = [names]

        for name in names if names else self.names:
            # calculate de-biasing factor
            n = self.psdm.psds[name].N_avg
            debias = n / (n + 1)

            if self.active_calibration and name is self.psdm.ex_axis:
                f_ex = self.psdm.get_ex_freq(unit='Hz')
                self.psdm.exclude_freq(f_ex, names=self.psdm.ex_axis)

            freq = self.get_freq(name, unit='Hz')
            psd = self.get_psd(name, unit='V**2/Hz')
            (f_c, D, chi2, bounds) = calc_anal_lsq_opt(freq, psd,
                                                       verbose=verbose,
                                                       **kwargs)
            # check for nans at the output of calc_anal_lsq_opt
            N_nans = sum(1 * isnan(val) for val in [f_c, D, chi2])
            if N_nans > 0:
                model = None
                f_c = 1000
                D = 1
                chi2 = 0
                bounds = (min(freq), max(freq))
                if verbose:
                    print('Analytical fit did not succeed. Using the '
                          'following default values: '
                          'f_c=1000 hz, D=1.0 V²/s, chi2=0, bounds={}'
                          ''.format(bounds))
            else:
                model = 'analytical least squares'

            keep = logical_and(bounds[0] <= freq, freq <= bounds[1])
            freq_keep = freq[keep]
            N_free = len(freq_keep) - 2
            lf_eval = lorentzian_psd(freq_keep, D * debias, f_c)

            FR = FitResult(name=name,
                           D=D * debias,
                           D_err=nan,
                           D_unit='V**2/s',
                           f_c=f_c,
                           f_c_err=nan,
                           freq=freq_keep,
                           freq_unit='Hz',
                           eval=lf_eval,
                           psd_unit='V**2/Hz',
                           residuals=None,
                           chi2=chi2,
                           redchi2=chi2/N_free,
                           nfree=N_free,
                           params=None,
                           minimizer=None,
                           model=model)

            self.anal_fits.update({name: FR})

    def plot_anal_fits(self, names=None, axis=None, **kwargs):
        """
        Plot the analytical least squares fits.

        Arguments
        ---------
        names : str
            specify which axis to plot. E.g. names='xy', would plot only x and
            y.
        axis : matplotlib:Axis
            Specify in which axis to add the plot
        **kwargs
            Keyword arguments handed over to *plot_fit()*.
        """
        if names and not isinstance(names, list):
            names = [names]

        for name in names if names else self.names:
            fig = self.anal_fits[name].plot_fit(axis=axis, **kwargs)
            if axis is None:
                axis = fig.axes[0]

        return fig

    def collective_psd_fit(self,
                           names=None,
                           model='lorentzian',
                           fitreport=False,
                           aliasing=False,
                           N_alias=9,
                           verbose=False,
                           **kwargs):
        """
        Fits the power spectra to the selected model function collectively,
        which is modified by a low-pass filter. The parameters alpha and
        f3dB of the low-pass filter are the same for all psds.
        This can be used to find alpha and f3dB.

        Arguments
        ---------
        names : str or list of str
            names of the PSDs to include in the collective fit, defaults to all
            available names.
        model : {'lorentzian', 'hydro'}
            Define the model function to be used.
        aliasing : bool
            Whether to include aliasing. If True, the sample frequency of the
            psdmeasurment is used.
        N_alias : int
            Number that defines how many ranges of f_sample should be taken
            into account. The default N=9 give a very good approximation with
            deviations of less than 0.5%
        verbose : bool
            Give more info.

        Keyword Arguments
        -----------------
        Keyword arguments are passed to make_collective_psd_fit. They can be
        initial guess values for the parameters f3dB and alpha and D and f_c
        where the axes for the latter two are specified by '_name',
        i.e. D_x=1 would specify the initial guess for the diffusion constant
        of the axis 'x'. Further, the suffixes '_min' and '_max' identify lower
        and upper boundaries for the fitting.
        """
        if model.lower() is 'hydro':
            expset = self.psdm.exp_setting
            h_kw = {'radius': expset.get_radius(unit='m'),
                    'height': expset.get_height(unit='m'),
                    'temp': expset.get_temp(unit='K'),
                    'rho': expset.get_density_particle(unit='kg/m**3'),
                    'density_med': expset.get_density_medium(unit='kg/m**3'),
                    'viscosity': expset.get_viscosity(unit='Pa*s')}
            kwargs.update(h_kw)

        freq_dict = OrderedDict()
        psd_dict = OrderedDict()
        err_dict = OrderedDict()

        fs_ = []

        if names and not isinstance(names, list):
            names = [names]

        names = names if names else self.names

        for name in names:
            freq_dict[name] = self.get_freq(name, unit='Hz')
            psd_dict[name] = self.get_psd(name, unit='V**2/Hz')
            err_dict[name] = self.get_psd_err(name, unit='V**2/Hz')
            if 'D_' + name not in kwargs:
                kwargs['D_' + name] = self.anal_fits[name].get_D(unit='V**2/s')
            if 'f_c_' + name not in kwargs:
                kwargs['f_c_' + name] = self.anal_fits[name].get_f_c(unit='Hz')
            fs_.append(self.psdm.psds[name].get_f_sample(unit='Hz'))

        if len(set(fs_)) > 1:
            warnings.warn('Modelling PSDs with different sampling rates '
                          'is not yet supported! Modelling aliasing will not '
                          'work correctly.')
        minimizer = make_collective_psd_fit(freq_dict,
                                            psd_dict,
                                            err_dict,
                                            model=model,
                                            aliasing=aliasing,
                                            f_sample=fs_[0],
                                            N_alias=9,
                                            verbose=verbose,
                                            **kwargs)
        params = minimizer.params

        model_ = collective_psd_fit_fun(params,
                                        names,
                                        freq_dict,
                                        minimizer.fun,
                                        minimizer.param_names)
        start = 0
        npara_per_axis = 2  # f_c and D
        ntotal = len(minimizer.residual)

        for name in names:
            freq = self.get_freq(name)
            len_ = len(freq)
            mod_ = model_[start:start+len_]
            res_ = minimizer.residual[start:start+len_]
            start += len_

            chi2 = (res_**2).sum()
            ndata = len_
            # f3dB and alpha are shared among all datapoints
            nshared = 2 * ndata / ntotal
            nfree = ndata - npara_per_axis - nshared
            redchi2 = chi2 / nfree

            fres = FitResult(name=name + ' (cltv)',
                             D=params['D_' + name].value,
                             D_err=params['D_' + name].stderr if params['D_' + name].stderr is not None else nan,
                             f_c=params['f_c_' + name].value,
                             f_c_err=params['f_c_' + name].stderr if params['f_c_' + name].stderr is not None else nan,
                             freq=freq,
                             eval=mod_,
                             residuals=res_,
                             chi2=chi2,
                             redchi2=redchi2,
                             conf_level=self.conf_level,
                             nfree=nfree,
                             params=params,
                             minimizer=minimizer,
                             model='collective lp-filtered ' + model +' psd fit'
                             )

            self.fits.update({name: fres})
            
        if fitreport and not verbose:
            report_fit(minimizer, show_correl=False)

    def setup_fit(self,
                  names=None,
                  model='hydro',
                  lp_filter=False,
                  lp_fixed=False,
                  aliasing=False,
                  f_sample=None,
                  N_alias=9,
                  debias=True,
                  f3dB=None,
                  alpha=None,
                  **fit_psd_kws):
        """
        Setup the psd fitting parameters.

        Arguments
        ---------
        names : str or list of str
            Defines the name(s) of the fits to be set up.
        model : {'lorentzian', 'hydro'}
            Define the model function to be used.
        lp_filter : bool
            If True the selected model will be modified by a low-pass
            filter.
        lp_fixed : bool
            If true the low pass filter parameters alpha and f3dB
            will be fixed. Expects the keywords 'alpha' and 'f3dB' and their
            respective values as kwargs. If they are not provided they are set
            to alpha = 0.5 and f3dB=10e3 Hz.
        aliasing : bool
            If aliasing should be considered or not.
        f_sample : float
            If aliasing is True, the sampling frequency of the original signal
            is needed to model the aliasing. If None, it is set to 2 * f_max,
            where f_max is the highest frequency in the frequency vector.
        N_alias : int
            Number that defines how many ranges of f_sample should be taken
            into account. The default N=9 give a very good approximation with
            deviations of less than 0.5%
        debias : bool
            Whether to account for biasing after the fitting. This only effects
            the Diffusion coefficient by a factor of N_avg / (N_avg + 1).
            N_avg is taken from the PSDMeasurement object.
        f3dB : float
            Cut-off frequency of the low-pass filter. Either the initial guess
            or the fixed value if lp_fixed is True.
        alpha : float
            Filter efficiency. Only a fraction (0 <= alpha <= 1) of the signal
            that is not low pass filtered. Thus, alpha = 1, will produce no
            low-pass filter at all, whereas alpha = 0, will produce a 1st-order
            low-pass filter.
        **fit_psd_kws : keyword arguments passed to
            pyotc.psd_fitting.fit_psd().
        """
        if names and not isinstance(names, list):
            names = [names]

        for name in names if names else self.names:
            kws = fit_psd_kws
            kws['model'] = model.lower()
            kws['lp_filter'] = lp_filter
            kws['lp_fixed'] = lp_fixed
            kws['aliasing'] = aliasing
            if aliasing and f_sample is None:
                try:
                    f_sample = self.psdm.psds[name].get_f_sample(unit='Hz')
                except:
                    pass
            kws['f_sample'] = f_sample
            kws['N_alias'] = N_alias
            kws['debias'] = debias
            if f3dB:
                kws['f3dB'] = f3dB
            if alpha:
                kws['alpha'] = alpha
            self.fit_kwargs[name] = kws

    def fit_psd(self,
                name,
                dynamic_bounds=False,
                bounds=None,
                f_exclude=None,
                use_heights=False,
                plot_fit=False,
                plot_axis=None,
                plot_masked_data=True,
                fitreport=False,
                verbose=False,
                **kwargs):
        """
        Fit psd of data 'name' to the selected model.

        Consider setting up the fit first by self.setup_fit().

        Arguments
        ---------
        name : str
            name of the axis to be fitted.
        dynamic_bounds : bool
            Whether to determine the frequency bounds from the analytical fit.
            If True the upper bound is set one order of magnitude
            higher than the analytically determined corner freuqency. The
            lower bound matches the optimal minimum frequency from the
            calc_anal_lsq_opt() function, if this frequency is less than one
            order of magnitude lower than the analytically determined corner
            freuqency.
        bounds : tuple (fmin, fmax) as limits for fitting
        f_exclude : float or list
            frequency or a list of frequencies that should be excluded from
            the fit.
        use_heights : bool
            Only important for hydrodyn. cor. fitting of a psd. If True, the
            internally set height is used otherwise it is set to infinity.
        plot_fit : bool
            Whether to plot the fit.
        plot_axis : matplotlib.Axis or None
        plot_masked : bool
            Whether to plot masked data as well.
        fitreport : bool
            Whether to print a report of the fit or not.
        verbose : bool
            Print details for debugging.

        Keyword Arguments
        -----------------
        height : float
            If model is 'hydro'
            Distance between the bead center and the surface,
            if it is not specified infinity is used.
        density_med : float
            Density of the medium in kg/m³. Only used if model is 'hydro'. If
            None, the density of water is calculated at the temperature
            specified in the measurement object.
        viscosity : float
            Viscosity of the medium. Only used if model is 'hydro'. If
            None, the viscosity of water is calculated at the temperature
            specified in the measurement object.

        Note
        ----
        Other Keyword arguments can also be initial guesses of 'D' and 'f_c',
        one can also set their limits by specifying '_min' and '_max' for the
        respective fitting parameter (see also: gen_fit_pars()).

        If model is 'hydro', the values of the particle radius, temperature and
        rho (the density of the particle) are taken from the PSDMeasurement-
        object.

        Setting bounds for a certain 'name' resets the former mask of the
        corresponding PSD-object. All other masked values at specific
        frequencies must be defined again via f_exclude.
        """
        if bounds or f_exclude:
            self.psdm.psds[name].reset_mask()

        if dynamic_bounds:
            anal_fit = self.anal_fits[name]
            f_c = anal_fit.f_c
            opt_fmin = anal_fit.get_freq_bounds()[0]
            # take f_min from analytical fit only if this bound is one
            # order of magnitude smaller than the corner frequency
            fmin = opt_fmin if opt_fmin < f_c / 10 else f_c / 10
            bounds = (fmin, 10 * f_c)

        if bounds:
            fmin, fmax = bounds
            self.psdm.exclude_freq_outside(fmin, fmax, names=name)
        if f_exclude:
            self.psdm.exclude_freq(f_exclude, names=name)

        if self.active_calibration and name == self.psdm.ex_axis:
            f_ex = self.psdm.get_ex_freq(unit='Hz')
            self.psdm.exclude_freq(f_ex, names=name)
            isexcited = True
        else:
            isexcited = False

        # create fit options if not done yet
        if name not in self.fit_kwargs:
            self.setup_fit(names=name)
            warnings.warn('Fit was not setup. Using default settings. '
                          'See setup_fit()')
        fitkwargs = self.fit_kwargs[name]

        kwargs.update(fitkwargs)

        # should the fit be de-biased?
        N_avg = self.psdm.psds[name].N_avg
        if kwargs.pop('debias'):
            deb = N_avg / (N_avg + 1)
        else:
            deb = 1

        # set initial values for D and f_c if not provided
        if 'D' not in kwargs:
            kwargs['D'] = self.anal_fits[name].D
        if 'f_c' not in kwargs:
            kwargs['f_c'] = self.anal_fits[name].f_c

        if kwargs['model'] == 'hydro':
            expset = self.psdm.exp_setting

            radius = expset.get_radius(unit='m')
            temp = expset.get_temp(unit='K')
            density_p = expset.get_density_particle(unit='kg/m**3')
            density_m = expset.get_density_medium(unit='kg/m**3')
            viscosity = expset.get_viscosity(unit=('Pa*s'))
            lateral = self.psdm.psds[name].is_lateral()

            if use_heights:
                height = kwargs['height'] = expset.get_height(unit='m')
            else:
                height = inf

            kwargs['radius'] = radius
            kwargs['temp'] = temp
            kwargs['rho'] = density_p
            kwargs['density_med'] = density_m
            kwargs['viscosity'] = viscosity
            kwargs['lateral'] = lateral

            s_pars = ('Radius = {0:1.3e} m\t height = {1:1.3e} m \t '
                      'T = {2:1.3f} K\t density = {3:1.3f}\n'
                      'density medium = {4:1.3f} kg/m**3 \t '
                      'viscosity = {5:1.3f} Pa*s'
                      ''.format(radius, height, temp, density_p,
                                density_m, viscosity))
        else:
            s_pars = ''

        lp_filter = kwargs['lp_filter']
        lp_fixed = kwargs['lp_fixed']

        if fitreport or verbose:
            print('----------')
            print(" '{0:s}' ".format(name).center(10))
            print('----------')

        freq = self.get_freq(name, unit='Hz')
        psd = self.get_psd(name, unit='V**2/Hz')
        err = self.get_psd_err(name, unit='V**2/Hz')

        modelfit = fit_psd(freq,
                           psd,
                           psd_err=err,
                           N_avg=N_avg,
                           verbose=verbose,
                           **kwargs)

        modelname = modelfit.model.name.replace('Model(', '').replace(')', '')
        if fitreport or verbose:
            print('Model: {0:s}'.format(modelname, name))
            print('Flags and parameters:')
            print('Low pass filter: {}\t fixed: {}'
                  ''.format(lp_filter, lp_fixed))
            if lp_filter:
                print('f_3dB = {0:1.1f}\t alpha = {1:1.3f}'
                      ''.format(kwargs['f3dB'], kwargs['alpha']))
            if s_pars:
                print(s_pars)
            print('De-bias factor = {0:1.3f}'.format(deb))

            if not verbose:
                report_fit(modelfit)

        params = modelfit.params
        FR = FitResult(name=name,
                       D=params['D'].value * deb,
                       D_err=params['D'].stderr * deb if params['D'].stderr is not None else nan,
                       D_unit='V**2/s',
                       f_c=params['f_c'].value,
                       f_c_err=params['f_c'].stderr if params['f_c'].stderr is not None else nan,
                       freq=freq,
                       freq_unit='Hz',
                       eval=1 / modelfit.eval() * deb,
                       psd_unit='V**2/Hz',
                       residuals=modelfit.residual,
                       chi2=modelfit.chisqr,
                       redchi2=modelfit.redchi,
                       conf_level=self.conf_level,
                       nfree=modelfit.nfree,
                       params=params,
                       minimizer=modelfit,
                       model=modelname
                       )
        self.fits[name] = FR

        # The result of excited axis can be accessed without knowing the name
        if isexcited:
            if 'excited_axis' in self.fits.keys():
                warnings.warn('Either the fit of the excited axis was repeated'
                              ' or there are more than one excited axis. '
                              'The "excited_axis" key is overwritten.'
                              'Calling the fit intrisically by "excited_axis" '
                              'and not by the name of the axis can lead to '
                              'unwanted results.')
            self.fits['excited_axis'] = FR

        if plot_fit:
            self.plot_fits(names=name,
                           plot_axis=plot_axis,
                           plot_masked_data=plot_masked_data)

    def fit_psds(self,
                 names=None,
                 dynamic_bounds=False,
                 bounds=None,
                 f_exclude=None,
                 use_heights=False,
                 calc_results=True,
                 plot_fits=False,
                 plot_axis=None,
                 fitreport=False,
                 verbose=False,
                 **kwargs):
        """
        Fit psds.

        Same method as fit_psd. this one iterates through the axes 'names' or
        all available axes in self.names.

        Arguments
        ---------
        names : str or list of str
            names to run the fiiting routine all available once are used if
            None.
        dynamic_bounds : bool
            Whether to determine the frequency bounds from the analytical fit.
            If True the upper bound is set one order of magnitude
            higher than the analytically determined corner freuqency. The
            lower bound matches the optimal minimum frequency from the
            calc_anal_lsq_opt() function, if this frequency is less than one
            order of magnitude lower than the analytically determined corner
            freuqency.
        bounds : dict
            Dictionary with names of the axes as keys and tuples of fmin and
            fmax as values.
        f_exclude : dict
            Dictionary with names of the axes as keys and lists of frequencies
            to be excluded.
        use_heights : bool
            Only important for hydrodyn. cor. fitting of a psd. If True, the
            internally set height is used otherwise it is set to infinity.
        calc_results : bool
            Calculate the pc and ac results if True.
        plot_fits : bool
            Whether to plot the fits.
        plot_axis : matplotlib.Axis or None
        fitreport : bool
            Whether to print a report of the fit or not.
        verbose : bool
            Print details for debugging.
        """
        if fitreport or verbose:
            print('========================================================')

        if names and not isinstance(names, list):
            names = [names]

        for name in names if names else self.names:
            if hasattr(bounds, 'keys'):
                try:
                    bnds = bounds[name]
                except:
                    bnds = None
            else:
                bnds = bounds

            if hasattr(f_exclude, 'keys'):
                try:
                    f_ex = f_exclude[name]
                except:
                    f_ex = None
            else:
                f_ex = f_exclude

            self.fit_psd(name=name,
                         dynamic_bounds=dynamic_bounds,
                         bounds=bnds,
                         f_exclude=f_ex,
                         use_heights=use_heights,
                         fitreport=fitreport,
                         verbose=verbose,
                         **kwargs)
            if calc_results:
                self.calc_pc_results(names=name, verbose=verbose)

        if plot_fits:
            self.plot_fits(names=names,
                           plot_axis=plot_axis)

        if fitreport or verbose:
            print('========================================================')

        if self.active_calibration:
            if all((name in self.fits) for name in self.names):
                self.calc_ac_results(verbose=verbose)

        if verbose:
            self.print_pc_results()
            self.print_ac_results()

        if plot_fits:
            plt.show()

    def plot_fits(self,
                  names=None,
                  plot_axis=None,
                  plot_data=True,
                  plot_masked_data=True,
                  save_plots=False,
                  save_as='png',
                  directory=None,
                  filename=None,
                  **kwargs):
        """
        Plots the fits.
        """
        if names and not isinstance(names, list):
            names = [names]

        for name in names if names else self.names:
            if plot_data:
                fig = self.psdm.plot_psds(names=name,
                                          axis=plot_axis,
                                          **kwargs)

                if plot_axis is None:
                    plot_axis = fig.axes[0]

                if plot_masked_data:
                    self.psdm.plot_psds(names=name,
                                        axis=plot_axis,
                                        plot_masked=True,
                                        **kwargs)

            fig = self.fits[name].plot_fit(axis=plot_axis, **kwargs)
            if plot_axis is None:
                plot_axis = fig.axes[0]

        if save_plots:
            directory = directory if directory else './'
            ptf = ''.join([directory, filename, '.', save_as])
            fig.patch.set_alpha(1.0)
            fig.patch.set_color('white')
            fig.savefig(ptf, dpi=150, format=save_as)

        return fig

    def is_outlier(self, names=None, conf_level=None):
        """
        Do the fits lie within the given confidence interval?
        """
        if not conf_level:
            conf_level = self.conf_level

        if names and not isinstance(names, list):
            names = [names]

        names = names if names else self.names

        return {name: self.fits[name].is_outlier(conf_level)
                for name in names}

    def get_basepower(self, unit='V**2'):
        """
        Determines the basepower and its error from the fit.
        The error is calculated and should be understood as an upper error
        estimate. A correct calculation should consider the covariances
        from the fitting parameters.
        As far as the excitation amplitude is at least two orders of magnitude
        higher as the basepower, the impact that this error has is of minor
        interest.
        """
        if unit != self._ex_power_unit:
            conv = ureg(self._ex_power_unit).to(unit).magnitude
        else:
            conv = 1.0

        ex_axis = self.psdm.ex_axis
        fit = self.fits[ex_axis]
        f_ex = self.psdm.get_ex_freq(unit='Hz')
        df = self.psdm.psds[ex_axis].df

        # evaluate the fit at the excitation frequency
        # take inverse because fit is done inversely
        bp = (1 / fit.minimizer.eval(f=array([f_ex]))[0]) * df * conv
        bp_err = bp / sqrt(self.psdm.psds[ex_axis].N_avg) * conv

        return (bp, bp_err)

    def calc_pc_results(self, names=None, verbose=False):
        """
        Calculate the passive calibration results.

        The following values are calculated:

         - displacement sensitivity
         - trap stiffness
         - Stokes drag

        the calculations assume Stokes' drag of a sphere of the given radius,
        viscosity and temperature.
        """
        T = self.psdm.exp_setting.get_temp(unit='K')
        dT = self.psdm.exp_setting.get_temp_err(unit='K')

        drag = self.psdm.get_stokes_drag(unit='N*s/m')
        ddrag = self.psdm.get_stokes_drag_err(unit='N*s/m')

        D_0 = (k_B * T / drag)  # Nm/K * K / (Ns/m) = m²/s
        dD_0 = (dT / T + ddrag / drag) * D_0

        if verbose:
            print('passive calibration results:')
            print('T = {0:1.3e} +/- {1:1.3e} K'.format(T, dT))
            print('Stokes drag = {0:1.3e} +/- {1:1.3e} Ns/m'
                  ''.format(drag, ddrag))
            print('D_0 = k_B * T / (6*pi*eta*R) = {0:1.3e} +/- {1:1.3e} m^2/s'
                  ''.format(D_0, dD_0))

        if names and not isinstance(names, list):
            names = [names]

        for name in names if names else self.names:
            if name not in self.fits.keys():
                raise Exception("Fit for '{}' not present."
                                "Try to fit the PSD first.".format(name))
            D_V = self.fits[name].D  # V**2/s
            dD_V = self.fits[name].D_err

            f_c = self.fits[name].f_c  # Hz
            df_c = self.fits[name].f_c_err

            # dissens
            beta_pc = (D_0 / D_V)**0.5 * 1e6  # m/V = 1e6 nm/mV
            dbeta_pc = (0.5 * dD_0 / D_0 + dD_V / D_V) * beta_pc

            # trap_stiffness
            # [kappa] = 1/s * Ns/m = N/m = 1e3 pN/nm
            kappa_pc = (2 * pi * f_c * drag) * 1e3
            dkappa_pc = (df_c / f_c + ddrag / drag) * kappa_pc

            pc_result = Result(dissens=beta_pc,
                               dissens_err=dbeta_pc,
                               dissens_unit='nm/mV',
                               trap_stiffness=kappa_pc,
                               trap_stiffness_err=dkappa_pc,
                               trap_stiffness_unit='pN/nm',
                               drag=drag * 1e9,  # Ns/m = 1e9 nNs/m
                               drag_err=ddrag * 1e9,
                               drag_unit='nN*s/m',
                               excited=False,
                               )
            self.pc_results.update({name: pc_result})

            if verbose:
                print(name)
                print('D_V = {0:1.3e} +/- {1:1.3e} V^2/s'.format(D_V, dD_V))
                print('f_c = {0:1.3e} +/- {1:1.3e} Hz'.format(f_c, df_c))
                print('beta_pc = {0:1.3e} +/- {1:1.3e} nm/mV'
                      ''.format(beta_pc, dbeta_pc))
                print('kappa_pc = {0:1.3e} +/- {1:1.3e} pN/nm'
                      ''.format(kappa_pc, dkappa_pc))

        if self.active_calibration and self.psdm.ex_axis in self.pc_results:
            self.pc_results['excited_axis'] = self.pc_results[self.psdm.ex_axis]

    def calc_ac_results(self, verbose=False):
        """
        Calculates the values:
         - displacement sensitivity
         - trap stiffness
         - drag
        for the given PSDM.
        """
        names = self.names.copy()
        for name in names:
            if name not in self.fits.keys():
                raise Exception("Fit for '{}' not present."
                                "Try to fit the PSD first.".format(name))
        # name of ac-excited PSD
        ex_axis = self.psdm.ex_axis
        # TODO axial ac-calibration
        if self.psdm.psds[ex_axis].direction is 'axial':
            warnings.warn('Axial excitation involves height dependent-changes '
                          'of the drag coefficient that are dependent on the '
                          'amplitude of excitation itself. These '
                          'dependencies are not yet accounted for. Thus '
                          'it is recommended to excite in the lateral '
                          'direction. Ignoring this you accept the risk of '
                          'large errors. In other words, be sure you really '
                          'know what you do.')
        # excited frequency
        f_ex = self.psdm.ex_freq
        # excitation amplitude in meters
        A = self.psdm.get_ex_amplitude(unit='m')
        dA = self.psdm.get_ex_amplitude_err(unit='m')
        # power at excited frequency in V**2/Hz
        P = self.psdm.get_ex_power(unit='V**2')
        dP = self.psdm.get_ex_power_err(unit='V**2')
        # thermal (base) power density at excited frequency
        base, dbase = self.get_basepower(unit='V**2')
        # corner frequency from fit to excited axis
        f_c = self.fits[ex_axis].f_c
        df_c = self.fits[ex_axis].f_c_err
        # theoretical power for that excited frequency in m²
        W_th = A**2 / (2 * (1 + (f_c/f_ex)**2))
        dW_th = (2 * (dA / A + df_c / f_c)) * W_th
        # actual (measured) power in V²
        W_exp = (P - base)
        dW_exp = (dP + dbase)

        if verbose:
            print('ex amp: {0:1.3e} +/- {1:1.3e} m'.format(A, dA))
            print('ex power: {0:1.3e} +/- {1:1.3e} V**2'
                  ''.format(P, dP))
            print('base power: '
                  '{0:1.3e} +/- {1:1.3e} V**2'.format(base, dbase))
            print('W_th: {0:1.3e} +/- {1:1.3e} m**2'.format(W_th, dW_th))
            print('W_exp: {0:1.3e} +/- {1:1.3e} V**2'.format(W_exp, dW_exp))

        # displacement sensitivity for the excited axis
        beta_ac = (W_th / W_exp)**0.5  # sqrt(m²/V²) = m/V
        dbeta_ac = (0.5 * (dW_th / W_th + dW_exp / W_exp)) * beta_ac
        # Temperature in Kelvin
        T = self.psdm.exp_setting.get_temp(unit='K')
        dT = self.psdm.exp_setting.get_temp_err(unit='K')
        # diffusion constant determined from fit (V²/s)
        D_ac = self.fits[ex_axis].D
        dD_ac = self.fits[ex_axis].D_err
        # drag in physical units Nm/K * K / (m²/V² * V²/s) = Ns/m
        gamma_ac = k_B * T / (beta_ac**2 * D_ac)
        dgamma_ac = (dT / T + 2 * dbeta_ac / beta_ac + dD_ac / D_ac) * gamma_ac
        # trap stiffness in physical units (N/m)
        kappa_ac = 2 * pi * f_c * gamma_ac
        dkappa_ac = (df_c / f_c + dgamma_ac / gamma_ac) * kappa_ac

        if verbose:
            print('beta_ac = {0:1.3e} +/- {1:1.3e} nm/mV'
                  ''.format(beta_ac * 1e6, dbeta_ac * 1e6))
            print('T = {0:1.3e} +/- {1:1.3e} K'.format(T, dT))
            print('D_ac = {0:1.3e} +/- {1:1.3e} V**2/s'.format(D_ac, dD_ac))
            print('gamma_ac = {0:1.3e} +/- {1:1.3e} nN*s/m'
                  ''.format(gamma_ac * 1e9, dgamma_ac * 1e9))
            print('kappa_ac = {0:1.3e} +/- {1:1.3e} pN/nm'
                  ''.format(kappa_ac * 1e3, dkappa_ac * 1e3))

        ac_result = Result(dissens=beta_ac * 1e6,
                           dissens_err=dbeta_ac * 1e6,
                           dissens_unit='nm/mV',
                           drag=gamma_ac * 1e9,
                           drag_err=dgamma_ac * 1e9,
                           drag_unit='nN*s/m',
                           trap_stiffness=kappa_ac * 1e3,
                           trap_stiffness_err=dkappa_ac * 1e3,
                           trap_stiffness_unit='pN/nm',
                           excited=True
                           )
        self.ac_results[ex_axis] = ac_result
        self.ac_results['excited_axis'] = ac_result

        names.remove(ex_axis)

        for name in names:
            D = self.fits[name].D  # V**2/s
            dD = self.fits[name].D_err
            f_c = self.fits[name].f_c  # Hz
            df_c = self.fits[name].f_c_err

            beta = beta_ac * sqrt(D_ac / D)  # m/V
            dbeta = (dbeta_ac / beta_ac + 0.5 * (dD_ac / D_ac + dD / D)) * beta

            kappa = 2 * pi * f_c * gamma_ac  # N/m
            dkappa = (df_c / f_c + dgamma_ac / gamma_ac) * kappa

            ac_result = Result(dissens=beta * 1e6,
                               dissens_err=dbeta * 1e6,
                               dissens_unit='nm/mV',
                               drag=gamma_ac * 1e9,
                               drag_err=dgamma_ac * 1e9,
                               drag_unit='nN*s/m',
                               trap_stiffness=kappa * 1e3,
                               trap_stiffness_err=dkappa * 1e3,
                               trap_stiffness_unit='pN/nm',
                               excited=False
                               )
            self.ac_results[name] = ac_result

    def print_ac_results(self):
        self.print_results(active_calibration=True)

    def print_pc_results(self):
        self.print_results(active_calibration=False)

    def print_results(self, active_calibration=True):
        """
        Print the results of the psd fits.
        """
        if active_calibration:
            if not self.active_calibration:
                print("PSD fit is set to 'passive calibration'. Check the "
                      "measurement-object's active_calibration attribute and "
                      "the ones starting with ex_.")
                return None
            s = 'ac_'
            s2 = 'Active calibration results'
        else:
            s = 'pc_'
            s2 = 'Passive calibration results'

        print('========================================================')
        print(s2)

        for name in self.names:
            # get the results of the passive or active calibration calculations
            result = getattr(self, s + 'results')[name]
            print('--------------------------------------------------------')
            print("'{}'".format(name))
            print('--------------------------------------------------------')
            print('Displacement sensitivity: {0:1.3e} +/- {1:1.3e} {2}'
                  ''.format(result.get_dissens(unit='nm/mV'),
                            result.get_dissens_err(unit='nm/mV'),
                            result.dissens_unit))
            print('Trap stiffness: {0:1.3e} +/- {1:1.3e} {2}'
                  ''.format(result.get_trap_stiffness(unit='pN/nm'),
                            result.get_trap_stiffness_err(unit='pN/nm'),
                            result.trap_stiffness_unit)
                  )
            print('Drag: {0:1.3e} +/- {1:1.3e} {2}'
                  ''.format(result.get_drag(unit='nN*s/m'),
                            result.get_drag_err(unit='nN*s/m'),
                            result.drag_unit))
        print('========================================================')

    def write_results_to_file(self, directory=None, fname=None, fext='.txt'):
        """
        Write fit results to a file.
        """
        if directory is None:
            try:
                directory = self.psdm.directory
            except:
                directory = './'
        if fname is None:
            try:
                fname = self.psdm.paramfile
            except:
                dstr = time.strftime("%Y-%m-%d_%H-%M")
                pfix = '_psd_fit_results'
                fname = pfix + dstr

        if not fname.endswith(fext):
            fname += fext

        pfile = join(directory, fname)

        cfg = ConfigParser()

        if isfile(pfile):
            cfg.read(pfile)

        for name in self.names:
            fit_results = OrderedDict()

            fit_results[co.model] = self.fits[name].model

            fit_results[co.f_c] = '{0:1.5e}'.format(self.fits[name].f_c)
            fit_results[co.df_c] = '{0:1.5e}'.format(self.fits[name].f_c_err)
            fit_results[co.f_c_unit] = self.fits[name].freq_unit

            fit_results[co.D] = '{0:1.5e}'.format(self.fits[name].D)
            fit_results[co.dD] = '{0:1.5e}'.format(self.fits[name].D_err)
            fit_results[co.D_unit] = self.fits[name].D_unit

            fit_results[co.bounds] = ', '.join(str(a) for a in self.fits[name].get_freq_bounds())
            fit_results[co.redchi2] = '{0:1.5e}'.format(self.fits[name].redchi2)
            fit_results[co.nfree] = self.fits[name].nfree
            fit_results[co.outlier] = self.fits[name].is_outlier()
            fit_results[co.conf_level] = self.fits[name].conf_level

            section = 'FIT_RESULTS_' + name.upper()
            if not section in cfg.sections():
                cfg.add_section(section)
            cfg[section] = fit_results

            # passive calibration results
            pc_results = OrderedDict()
            keys = ['dissens', 'dissens_err', 'dissens_unit',
                    'trap_stiffness', 'trap_stiffness_err',
                    'trap_stiffness_unit',
                    'drag', 'drag_err', 'drag_unit']
            for key in keys:
                key_name = getattr(co, key)
                pc_results[key_name] = getattr(self.pc_results[name], key)

            section = 'PASSIVE_CALIBRATION_RESULTS_' + name.upper()
            if not section in cfg.sections():
                cfg.add_section(section)
            cfg[section] = pc_results

            # active calibration results
            if self.active_calibration:
                ac_results = OrderedDict()

                for key in keys:
                    ac_results[key] = getattr(self.ac_results[name], key)
                section = 'ACTIVE_CALIBRATION_RESULTS_' + name.upper()
                if not section in cfg.sections():
                    cfg.add_section(section)
                cfg[section] = ac_results
        try:
            with open(pfile, 'w') as cfl:
                cfg.write(cfl)
        except:
            raise Exception('Unable to write results to file {}'
                            ''.format(pfile))
        print(pfile)
